/*****************************************************************************\
* Copyright (c) 2002 Pelle Johansson.                                         *
* All rights reserved.                                                        *
*                                                                             *
* This file is part of the moftpd package. Use and distribution of            *
* this software is governed by the terms in the file LICENCE, which           *
* should have come with this package.                                         *
\*****************************************************************************/

/* $moftpd: commands.c,v 1.46.2.6 2004/03/16 20:07:26 morth Exp $ */
/* $Id$ */

#include "commands.h"

#include "connection.h"
#include "main.h"
#include "user.h"
#include "utf8fs/file.h"
#include "defaults.h"
#include "utf8fs/memory.h"

#include "system.h"

typedef struct command
{
  const char *name;
  int (*handler)(connection_t *conn, const char *arg, int expected);
  int unauthed, whileWorking;
} command_t;

const command_t commands[] =
{
  {"ABOR", command_abor, 1, 1},
  {"ACCT", command_acct, 0, 0},
  {"ADAT", command_adat, 1, 0},
  {"ALLO", command_allo, 0, 0},
  {"APPE", command_appe, 0, 0},
  {"AUTH", command_auth, 1, 0},
  {"CCC" , command_ccc , 1, 0},
  {"CDUP", command_cdup, 0, 0},
  {"CONF", command_conf, 1, 0},
  {"CWD" , command_cwd , 0, 0},
  {"DELE", command_dele, 0, 0},
  {"ENC" , command_enc , 1, 0},
  {"EPRT", command_eprt, 0, 0},
  {"EPSV", command_epsv, 0, 0},
  {"FEAT", command_feat, 1, 0},
  {"HELP", command_help, 1, 0},
  {"LANG", command_lang, 1, 0},
  {"LIST", command_list, 0, 0},
  {"LPRT", command_lprt, 0, 0},
  {"LPSV", command_lpsv, 0, 0},
  {"MIC" , command_mic , 1, 0},
  {"MKD" , command_mkd , 0, 0},
  {"MDTM", command_mdtm, 0, 0},
  {"MLST", command_mlst, 0, 0},
  {"MLSD", command_mlsd, 0, 0},
  {"MODE", command_mode, 0, 0},
  {"NLST", command_nlst, 0, 0},
  {"NOOP", command_noop, 1, 0},
  {"OPTS", command_opts, 0, 0},
  {"PASS", command_pass, 1, 0},
  {"QUIT", command_quit, 1, 1},
  {"PASV", command_pasv, 0, 0},
  {"PBSZ", command_pbsz, 0, 0},
  {"PORT", command_port, 0, 0},
  {"PROT", command_prot, 0, 0},
  {"PWD" , command_pwd , 0, 0},
  {"REIN", command_rein, 1, 0},
  {"REST", command_rest, 0, 0},
  {"RETR", command_retr, 0, 0},
  {"RMD" , command_rmd , 0, 0},
  {"RNFR", command_rnfr, 0, 0},
  {"RNTO", command_rnto, 0, 0},
  {"SITE", command_site, 0, 0},
  {"SIZE", command_size, 0, 0},
  {"SMNT", command_smnt, 0, 0},
  {"STAT", command_stat, 0, 1},
  {"STOR", command_stor, 0, 0},
  {"STOU", command_stou, 0, 0},
  {"STRU", command_stru, 0, 0},
  {"SYST", command_syst, 0, 0},
  {"TYPE", command_type, 0, 0},
  {"USER", command_user, 1, 0},
  {NULL}
};

extern const char *localeDir;
extern const char *localeSuffix;
extern unsigned long long maxMmapSize;
#ifdef HAVE_LIBSSL
extern SSL_CTX *sslCtx;
extern char *sslCertsPath;
#endif

int handle_command(connection_t *conn, char *line)
{
  const char *arg;
  const command_t *cmd;
  int expected;
  const char *esep, *enext;
  
  if(!strlen(line))
    return 0;
  
  /* Set up uid, gid, faked chroot, etc. */
  if(conn->user)
    user_setup_environ (conn->user, conn->authed);
  else
    drop_privs();
  set_locale (conn->currLang);
  
  if(conn->spontQuit && !conn->working)
  {
    if(conn->spontReply)
    {
      send_telnet(conn, conn->spontReply, 1);
      pfree(conn->spontReply, conn);
      conn->spontReply = NULL;
    }
    disconnected(conn);
    return 1;
  }
  
  /*
   * Try to cd back into the directory the connection is currently in. If it
   * fails, try the root directory before giving up. set_cwd() will make sure
   * we're not outside a faked chroot.
   */
  if ((!conn->cwd || set_cwd (conn->cwd, NULL)) && set_cwd ("/", NULL))
  {
    reply(conn, "421 Root directory inaccessable.");
    disconnected(conn);
    return 1;
  }
  
  arg = strchr(line, ' ');
  if(arg)
    *(char*)arg++ = 0;
  else
    arg = "";
  
  if(strlen(line) < 3 || strlen(line) > 4)
  {
    reply(conn, "500 Syntax error.");
    return 0;
  }
  
  expected = 0;
  for(esep = conn->expect; esep && !expected; esep = enext)
  {
    enext = strchr(esep, '|');
    if(enext)
      expected = !strncasecmp(line, esep, enext++ - esep);
    else
      expected = !strcasecmp(line, esep);
  }
  conn->expect = NULL;
  
  for(cmd = commands; cmd->name; cmd++)
  {
    if(!strcasecmp(cmd->name, line))
    {
      if(conn->working && !cmd->whileWorking)
      {
	/*
	 * rfc0959 is unclear of what to do when receiving commands during a
	 * 1xx command. In one place it says to queue them, but in others it
	 * refers to commands received during transfer. It does however state
	 * that it is illegal for the client to send commands while waiting
	 * for a reply (again, this seems to have some exceptions), so we
	 * mark the commands that are exceptions and deny the rest.
	 * This should really be a 4xx command, but a fitting such does not
	 * exist.
	 */
	reply(conn, "500 Please wait until the server is ready.");
      }
      else if(!conn->authed && !cmd->unauthed)
	reply(conn, "503 Please login.");
      else if(cmd->handler(conn, arg, expected))
	return 1;
      break;
    }
  }
  if(!cmd->name)
    reply(conn, "500 No such command %s.", line);
  
  if(conn->spontQuit && !conn->working)
  {
    disconnected(conn);
    return 1;
  }
  
  /* Store the current directory of this connection. */
  pfree (conn->cwd, conn);
  line = getcwd (NULL, MAXPATHLEN);
  if (line)
  {
    conn->cwd = pstring (line, conn);
    free (line);
  }
  else
    conn->cwd = NULL;
  
  return 0;
}

int check_prot_level (connection_t *conn, const char *path)
{
  switch (check_access (path, acEncrypted))
  {
  case 1:
    if (!(conn->protLevel & acEncrypted))
    {
      reply (conn, "534 Data connection must be encrypted.");
      return 1;
    }
    break;
  case -1:
    if (conn->protLevel & acEncrypted)
    {
      reply (conn, "534 Data connection must not be encrypted.");
      return 1;
    }
    break;
  }
  switch (check_access (path, acSigned))
  {
  case 1:
    if (!(conn->protLevel & acSigned))
    {
      reply (conn, "534 Data connection must be signed.");
      return 1;
    }
    break;
  case -1:
    if (conn->protLevel & acSigned)
    {
      reply (conn, "534 Data connection must not be signed.");
      return 1;
    }
    break;
  }
  return 0;
}

long long text_type_restart (const char *path, unsigned long long restart)
{
  int fd = open_file_reading (path, 0);
  char buf[1024], *bp;
  unsigned long long off;
  int l;
  
  if (fd < 0)
    return -1;
  
  off = 0;
  while (off < restart && (l = read (fd, buf, off - restart < 1024?
	off - restart : 1024)) > 0)
  {
    off += l;
    bp = buf;
    while ((bp = memchr (bp, '\n', l)))
    {
      restart--;
      l -= ++bp - buf;
    }
  }
  
  close (fd);
  return restart;
}


/* Login commands */

int command_user(connection_t *conn, const char *arg, int expected)
{
  if (conn->user && !strcmp (arg, conn->user->name))
  {
    reply (conn, "230 Already logged in as %s.", conn->user->name);
    return 0;
  }
  
  if (
#ifdef HAVE_LIBSSL
      (conn->ssl && conn->server->loginTLS && conn->server->tlsNoLogout) ||
#endif
      super_privs (1))
  {
    reply(conn, "503 Can't relogin.");
    return 0;
  }
  
  unauthenticated(conn);
#ifdef HAVE_LIBSSL
  conn->user = find_user (conn->server, arg, conn->ssl? 1 : 0, conn);
  if (conn->ssl && conn->user && !conn->user->allowSecLogin)
  {
    pfree (conn->user, conn);
    conn->user = NULL;
  }
#else
  conn->user = find_user (conn->server, arg, 0, conn);
#endif
  if(conn->user && !conn->user->passwordNeeded)
  {
    authenticated(conn);
    user_setup_environ (conn->user, 1);
    
    if(conn->user->anonymous)
      reply(conn, "230 Anonymous access granted.");
    else
      reply(conn, "230 Logged in as %s.", conn->user->name);
  }
  else if(!conn->user && !conn->server->passIfInvalid)
  {
    syslog (LOG_NOTICE, "%d: Unknown user %s.", conn->sock, arg);
    syslog (LOG_AUTHPRIV | LOG_NOTICE, "%d: Unknown user %s.",
	  conn->sock, arg);
    reply(conn, "530 %s", conn->server->userInvalidMsg?
	  conn->server->userInvalidMsg : defUserInvalidMsg);
    if (conn->server->sleepOnFail)
    {
      time (&conn->sleepUntil);
      conn->sleepUntil += conn->server->sleepOnFail;
    }
  }
  else
  {
    if (!conn->user)
    {
      syslog (LOG_NOTICE, "%d: Unknwon user %s.", conn->sock, arg);
      syslog (LOG_AUTHPRIV | LOG_NOTICE, "%d: Unknown user %s.",
	    conn->sock, arg);
    }
    if(conn->user && conn->user->anonymous)
      reply(conn, "331 %s", conn->server->anonPassMsg? conn->server->anonPassMsg
	    : defAnonPassMsg);
    else
      reply(conn, "331 %s", conn->server->passRequestMsg?
	    conn->server->passRequestMsg : defPassRequestMsg);
    conn->expect = "PASS";
  }
  return 0;
}

int command_pass(connection_t *conn, const char *arg, int expected)
{
  const char *pass;
  
  if(!expected)
  {
    reply(conn, "503 PASS not expected here.");
    return 0;
  }
  
  if(conn->user && conn->user->anonymous)
  {
    conn->email = pstring (arg, conn);
    authenticated (conn);
    user_setup_environ (conn->user, 1);
    reply(conn, "230 Anonymous access granted.");
  }
  else
  {
    pass = user_pass(conn->user);
    
    if(pass && !strcmp(crypt(arg, pass), pass))
    {
      authenticated(conn);
      user_setup_environ (conn->user, 1);
      reply(conn, "230 Logged in as %s.", conn->user->name);
    }
    else
    {
      if (conn->user)
      {
	syslog (LOG_NOTICE, "%d: User %s failed login.", conn->sock,
	      conn->user->name);
	syslog (LOG_AUTHPRIV | LOG_NOTICE, "%d: User %s failed login.",
	      conn->sock, conn->user->name);
      }
      if (conn->server->maxLoginAttempts && ++conn->failedLogins >=
	    conn->server->maxLoginAttempts)
      {
	reply (conn, "421 %s", conn->server->loginFailedMsg?
	      conn->server->loginFailedMsg : defLoginFailedMsg);
	disconnected (conn);
	return 1;
      }
      unauthenticated(conn);
      reply(conn, "530 %s", conn->server->loginFailedMsg?
	    conn->server->loginFailedMsg : defLoginFailedMsg);
      if (conn->server->sleepOnFail)
      {
	time (&conn->sleepUntil);
	conn->sleepUntil += conn->server->sleepOnFail;
      }
    }
  }
  return 0;
}

int command_acct(connection_t *conn, const char *arg, int expected)
{
  reply(conn, "202 ACCT not needed here.");
  return 0;
}

/* Logout commands */

int command_rein(connection_t *conn, const char *arg, int expected)
{
  if (super_privs (1))
  {
    reply(conn, "503 Can't relogin.");
    return 0;
  }
  
  if (conn->user)
    syslog (LOG_INFO, "%d: User %s logged out", conn->sock, conn->user->name);
  unauthenticated(conn);
  pfree (conn->currLang, conn);
  conn->currLang = NULL;
  if (conn->langFd >= 0)
    close_shared (conn->langFd);
  conn->langFd = -1;
  conn->protLevel = 0;
#ifdef HAVE_LIBSSL
  if (conn->ssl)
  {
    SSL_shutdown (conn->ssl);
    conn->sslState = 2;
  }
  else
#endif
    reply(conn, "220 You have been logged out.");
  return 0;
}

int command_quit(connection_t *conn, const char *arg, int expected)
{
  if (conn->working)
    reply_spont(conn, "221 Good bye.", 1);
  else
  {
    reply(conn, "221 Good bye.");
    disconnected(conn);
  }
  return 1;
}

/* Navigation commands */

int command_cwd(connection_t *conn, const char *arg, int expected)
{
  if(!strlen(arg))
    reply(conn, "501 Path missing.");
  else
  {
    if(set_cwd(chroot_path(arg, conn->user ? conn->user->home : NULL), NULL))
      reply(conn, "550 %s.", strerror(errno));
    else
      reply(conn, "250 \"%s\" is current directory.", print_path(chroot_path(
	    ".", NULL)));
  }
  return 0;
}

int command_cdup(connection_t *conn, const char *arg, int expected)
{
  return command_cwd(conn, "..", 0);
}

int command_smnt(connection_t *conn, const char *arg, int expected)
{
  /*
   * I suppose things like different drive letters in Windows and similar
   * systems should go here, but I've never seen any client implementing this
   * command.
   * Make SMNT <device> an alias for CWD <mount point> perhaps?
   */
  reply(conn, "502 Only one structure on this host.");
  return 0;
}

/* Transfer parameters */

int command_port(connection_t *conn, const char *arg, int expected)
{
  int a1, a2, a3, a4, p1, p2;
  unsigned int addr, port;
  struct sockaddr_in *addr4;
  
  if(conn->epsvOnly)
  {
    reply(conn, "503 You've chosen to disable the PORT command, use EPSV.");
    return 0;
  }
  
  switch(sscanf(arg, "%d,%d,%d,%d,%d,%d", &a1, &a2, &a3, &a4, &p1, &p2))
  {
  case 6:
    if(a1 < 0 || a1 > 255 || a2 < 0 || a2 > 255 || a3 < 0 || a3 > 255 ||
	  a4 < 0 || a4 > 255 || p1 < 0 || p1 > 255 || p2 < 0 || p2 > 255)
    {
      reply(conn, "501 Invalid port specification.");
      return 0;
    }
    addr = htonl(a1 << 24 | a2 << 16 | a3 << 8 | a4);
    port = p1 << 8 | p2;
    if(!conn->server->allowLowPorts && port < 1024)
    {
      syslog (LOG_NOTICE, "%d: PORT with low port: %d", conn->sock, port);
      reply(conn, "501 Port < 1024 not allowed.");
      return 0;
    }
    port = htons(port);
    addr4 = (struct sockaddr_in*)&conn->rDataAddr;
    if(!conn->server->allowForeign)
    {
      if(addr4->sin_family != AF_INET || addr4->sin_addr.s_addr != addr)
      {
	syslog (LOG_NOTICE, "%d: PORT with foreign address", conn->sock);
	reply(conn, "501 Foreign address not allowed.");
	return 0;
      }
    }
    addr4->sin_family = AF_INET;
#ifdef HAVE_STRUCT_SOCKADDR_IN_SIN_LEN
    addr4->sin_len = sizeof(struct sockaddr_in);
#endif
    addr4->sin_port = port;
    addr4->sin_addr.s_addr = addr;
    conn->passive = 0;
    reply(conn, "200 Port set successfully.");
    break;
  default:
    reply(conn, "501 Invalid port specification.");
    break;
  }
  return 0;
}

int command_lprt (connection_t *conn, const char *arg, int expected)
{
  int a[16], p[2];
  unsigned char ca[16], cp[2];
  int i, prt, al, pl;
  struct sockaddr_in addr4;
  struct sockaddr_in6 addr6;
  char buf[1024];
  
  if (conn->epsvOnly)
  {
    reply (conn, "503 You've chosen to disable the LPRT command, use EPSV.");
    return 0;
  }
  
  if (sscanf (arg, "%d,%d,%s", &prt, &al, buf) == 3)
  {
    if (prt != 4 && prt != 6)
    {
      reply (conn, "521 Supported address families are (4,6)");
      return 0;
    }
    if ((prt == 4 && al != 4) || (prt == 6 && al != 16))
    {
      reply (conn, "500 Invalid address length.");
      return 0;
    }
    for (i = 0; i < al; i++)
    {
      if (sscanf (buf, "%d,%s", a + i, buf) != 2)
      {
	reply (conn, "501 Number of fields doesn't match length given.");
	return 0;
      }
      if (a[i] < 0 || a[i] > 255)
      {
	reply (conn, "500 Invalid long port specification.");
	return 0;
      }
      ca[i] = a[i];
    }
    if (sscanf (buf, "%d,%s", &pl, buf) == 2)
    {
      if (pl != 2)
      {
	reply (conn, "500 Invalid port length.");
	return 0;
      }
      if (sscanf (buf, "%d,%d", &p[0], &p[1]) != 2)
      {
	reply (conn, "501 Number of fields doesn't match length given.");
	return 0;
      }
      cp[0] = p[0];
      cp[1] = p[1];
      
      pl = ntohs (*(unsigned short*)cp);
      if(!conn->server->allowLowPorts && pl < 1024)
      {
	syslog (LOG_NOTICE, "%d: LPRT with low port: %d", conn->sock, prt);
	reply(conn, "501 Port < 1024 not allowed.");
	return 0;
      }
      switch (prt)
      {
      case 4:
	addr4.sin_family = AF_INET;
#ifdef HAVE_STRUCT_SOCKADDR_IN_SIN_LEN
	addr4.sin_len = sizeof (struct sockaddr_in);
#endif
	addr4.sin_port = *(unsigned short*)cp;
	addr4.sin_addr.s_addr = *(unsigned int*)ca;
	if (!conn->server->allowForeign && !same_addr ((struct sockaddr*)&conn->
		 rDataAddr, (struct sockaddr*)&addr4, 0))
	{
	  syslog (LOG_NOTICE, "%d: LPRT with foreign address", conn->sock);
	  reply(conn, "501 Foreign address not allowed.");
	  return 0;
	}
	memcpy (&conn->rDataAddr, &addr4, sizeof (struct sockaddr_in));
	reply (conn, "200 Port set successfully.");
	return 0;
      case 6:
	addr6.sin6_family = AF_INET6;
#ifdef HAVE_STRUCT_SOCKADDR_IN_SIN_LEN
	addr6.sin6_len = sizeof (struct sockaddr_in6);
#endif
	addr6.sin6_port = *(unsigned short*)cp;
	memcpy (&addr6.sin6_addr, ca, 16);
	if (!conn->server->allowForeign && !same_addr ((struct sockaddr*)&conn->
		 rDataAddr, (struct sockaddr*)&addr6, 0))
	{
	  syslog (LOG_NOTICE, "%d: LPRT with foreign address", conn->sock);
	  reply(conn, "501 Foreign address not allowed.");
	  return 0;
	}
	memcpy (&conn->rDataAddr, &addr6, sizeof (struct sockaddr_in6));
	reply (conn, "200 Port set successfully.");
	return 0;
      }
    }
  }
  reply (conn, "500 Invalid long port specification.");
  return 0;
}

int command_eprt(connection_t *conn, const char *arg, int expected)
{
  unsigned char d = *arg++;
  char *narg;
  int prt;
  const char *addr, *port;
  struct addrinfo hints = {}, *res;
  
  if(conn->epsvOnly)
  {
    reply(conn, "503 You've chosen to disable the EPRT command, use EPSV.");
    return 0;
  }
  
  if(d < 33 || d > 126)
  {
    if (d)
      syslog (LOG_NOTICE, "%d: EPRT with invalid delimeter: %d", conn->sock, d);
    reply(conn, "501 Delimeter missing or invalid.");
    return 0;
  }
  
  narg = strchr(arg, d);
  if(!narg)
  {
    reply(conn, "501 Protocol field missing.");
    return 0;
  }
  *narg++ = 0;
  prt = atoi(arg);
  if(prt < 1 || prt > 2)
  {
    syslog (LOG_NOTICE, "%d: EPRT with invalid protocol: %d", conn->sock, prt);
    reply(conn, "522 Invalid protocol, use (1,2)");
    return 0;
  }
  arg = narg;
  narg = strchr(arg, d);
  if(!narg)
  {
    reply(conn, "501 Address field missing.");
    return 0;
  }
  *narg++ = 0;
  addr = arg;
  arg = narg;
  narg = strchr(arg, d);
  if(!narg)
  {
    reply(conn, "501 Port field missing.");
    return 0;
  }
  *narg++ = 0;
  port = arg;
  
  hints.ai_flags = AI_NUMERICHOST;
  hints.ai_socktype = SOCK_STREAM;
  switch(prt)
  {
  case 1:
    hints.ai_family = AF_INET;
    break;
  case 2:
    hints.ai_family = AF_INET6;
    break;
  }
  if(getaddrinfo(addr, port, &hints, &res) || !res)
  {
    reply(conn, "501 Invalid address specification.");
    return 0;
  }
  switch(res->ai_family)
  {
  case AF_INET:
    prt = ntohs(((struct sockaddr_in*)res->ai_addr)->sin_port);
    break;
  case AF_INET6:
    prt = ntohs(((struct sockaddr_in6*)res->ai_addr)->sin6_port);
    break;
  }
  if(!conn->server->allowForeign && !same_addr((struct sockaddr*)&conn->
	rDataAddr, res->ai_addr, 0))
  {
    syslog (LOG_NOTICE, "%d: EPRT with foreign address", conn->sock);
    reply(conn, "501 Foreign address not allowed.");
    return 0;
  }
  if(!conn->server->allowLowPorts && prt < 1024)
  {
    syslog (LOG_NOTICE, "%d: EPRT with low port: %d", conn->sock, prt);
    reply(conn, "501 Port < 1024 not allowed.");
    return 0;
  }
  
  memcpy(&conn->rDataAddr, res->ai_addr, res->ai_addrlen);
  conn->passive = 0;
  freeaddrinfo(res);
  reply(conn, "200 Port set successfully.");
  return 0;
}

int command_pasv(connection_t *conn, const char *arg, int expected)
{
  int hport;
  unsigned char *ap;
  
  if(conn->epsvOnly)
  {
    reply(conn, "503 You've chosen to disable the PASV command, use EPSV.");
    return 0;
  }
  
  if(conn->lDataAddr.ss_family != AF_INET)
  {
    reply(conn, "500 Not an Internet version 4 connection, use EPSV.");
    return 0;
  }
  
  conn->passive = 0;
  conn->passiveAccepted = 0;
  
  if(conn->dataSock >= 0)
  {
    /*
     * We know there's no transfer in progress, so safe to assume we can
     * silently close this.
     */
    close(conn->dataSock);
    remove_read_fd(conn->dataSock);
    remove_write_fd(conn->dataSock);
    conn->dataSock = -1;
  }
  
  conn->dataSock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if(conn->dataSock < 0)
  {
    syslog (LOG_ERR, "%d: Failed to create passive socket: %m", conn->sock);
    reply(conn, "425 Failed to create socket: %s.", strerror(errno));
    return 0;
  }
  for(hport = conn->server->minPasvPort; hport <= conn->server->maxPasvPort;
      hport++)
  {
    ((struct sockaddr_in*)&conn->lDataAddr)->sin_port = htons(hport);
    
    if(!bind(conn->dataSock, (struct sockaddr*)&conn->lDataAddr,
	  sizeof(struct sockaddr_in)))
      break;
  }
  
  if(hport > conn->server->maxPasvPort)
  {
    syslog (LOG_ERR, "%d: Failed to bind passive socket: %m", conn->sock);
    reply(conn, "425 Failed to bind socket: %s.", strerror(errno));
    close(conn->dataSock);
    conn->dataSock = -1;
    return 0;
  }
  
  if(listen(conn->dataSock, 1))
  {
    syslog (LOG_ERR, "%d: Failed to listen on passive socket: %m", conn->sock);
    reply(conn, "425 Failed to listen on socket: %s.", strerror(errno));
    close(conn->dataSock);
    conn->dataSock = -1;
    return 0;
  }
  
  add_read_fd(conn->dataSock, passive_acceptor, conn);
  conn->passive = 1;
  
  ap = (unsigned char*)&((struct sockaddr_in*)&conn->lDataAddr)->sin_addr;
  reply(conn, "227 Entering passive mode (%u,%u,%u,%u,%u,%u)", ap[0], ap[1],
	ap[2], ap[3], hport >> 8 & 0xFF, hport & 0xFF);
  return 0;
}

int command_lpsv (connection_t *conn, const char *arg, int expected)
{
  in_port_t *port;
  int hport, al;
  unsigned char *ca, *cp;
  
  if(conn->epsvOnly)
  {
    reply(conn, "503 You've chosen to disable the LPSV command, use EPSV.");
    return 0;
  }
  
  if (conn->dataSock >= 0)
  {
    /*
     * We know there's no transfer in progress, so safe to assume we can
     * silently close this.
     */
    close (conn->dataSock);
    remove_read_fd (conn->dataSock);
    remove_write_fd (conn->dataSock);
    conn->dataSock = -1;
  }
  
  conn->passive = 0;
  conn->passiveAccepted = 0;
  
  switch (conn->lDataAddr.ss_family)
  {
  case AF_INET:
    port = &((struct sockaddr_in*)&conn->lDataAddr)->sin_port;
    al = sizeof (struct sockaddr_in);
    break;
  case AF_INET6:
    port = &((struct sockaddr_in6*)&conn->lDataAddr)->sin6_port;
    al = sizeof (struct sockaddr_in6);
    break;
  default:
    reply(conn, "500 Unknown address family.");
    return 0;
  }
  
  conn->dataSock = socket (conn->lDataAddr.ss_family, SOCK_STREAM, IPPROTO_TCP);
  if (conn->dataSock < 0)
  {
    syslog (LOG_ERR, "%d: Failed to create passive socket: %m", conn->sock);
    reply (conn, "425 Failed to create socket: %s.", strerror (errno));
    return 0;
  }
  for (hport = conn->server->minPasvPort; hport <= conn->server->maxPasvPort;
       hport++)
  {
    *port = htons (hport);
    
    if(!bind (conn->dataSock, (struct sockaddr*)&conn->lDataAddr, al))
      break;
  }
  
  if (hport > conn->server->maxPasvPort)
  {
    syslog (LOG_ERR, "%d: Failed to bind passive socket: %m", conn->sock);
    reply (conn, "425 Failed to bind socket: %s.", strerror (errno));
    close (conn->dataSock);
    conn->dataSock = -1;
    return 0;
  }
  
  if (listen (conn->dataSock, 1))
  {
    syslog (LOG_ERR, "%d: Failed to listen on passive socket: %m", conn->sock);
    reply (conn, "425 Failed to listen on socket: %s.", strerror(errno));
    close (conn->dataSock);
    conn->dataSock = -1;
    return 0;
  }
  
  conn->passive = 1;
  add_read_fd (conn->dataSock, passive_acceptor, conn);
  
  cp = (unsigned char*)port;
  switch (conn->lDataAddr.ss_family)
  {
  case AF_INET:
    ca = (unsigned char*)&((struct sockaddr_in*)&conn->lDataAddr)->sin_addr;
    reply (conn, "228 Entering Long Passive Mode (4,4,%d,%d,%d,%d,2,%d,%d)",
	  ca[0], ca[1], ca[2], ca[3], cp[0], cp[1]);
    break;
  case AF_INET6:
    ca = (unsigned char*)&((struct sockaddr_in6*)&conn->lDataAddr)->sin6_addr;
    // Breaking a reply format makes it broken in the locale/messages file.
    reply (conn, "228 Entering Long Passive Mode (6,16,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,2,%d,%d)",
	  ca[0], ca[1], ca[2], ca[3],
	  ca[4], ca[5], ca[6], ca[7], ca[8], ca[9], ca[10], ca[11], ca[12],
	  ca[13], ca[14], ca[15], cp[0], cp[1]);
    break;
  }
  return 0;
}

int command_epsv(connection_t *conn, const char *arg, int expected)
{
  in_port_t *port;
  int hport, al;
  
  if(!strcasecmp(arg, "ALL"))
  {
    if(conn->epsvOnly)
      reply(conn, "200 Already in EPSV ALL mode.");
    else
    {
      conn->epsvOnly = 1;
      reply(conn, "200 EPSV ALL mode entered.");
    }
    return 0;
  }
  
  if(conn->dataSock >= 0)
  {
    /*
     * We know there's no transfer in progress, so safe to assume we can
     * silently close this.
     */
    close(conn->dataSock);
    remove_read_fd(conn->dataSock);
    remove_write_fd(conn->dataSock);
    conn->dataSock = -1;
  }
  
  conn->passive = 0;
  conn->passiveAccepted = 0;
  
  switch(conn->lDataAddr.ss_family)
  {
  case AF_INET:
    port = &((struct sockaddr_in*)&conn->lDataAddr)->sin_port;
    al = sizeof(struct sockaddr_in);
    break;
  case AF_INET6:
    port = &((struct sockaddr_in6*)&conn->lDataAddr)->sin6_port;
    al = sizeof(struct sockaddr_in6);
    break;
  default:
    reply(conn, "500 Unknown address family.");
    return 0;
  }
  
  conn->dataSock = socket(conn->lDataAddr.ss_family, SOCK_STREAM, IPPROTO_TCP);
  if(conn->dataSock < 0)
  {
    syslog (LOG_ERR, "%d: Failed to create passive socket: %m", conn->sock);
    reply(conn, "425 Failed to create socket: %s.", strerror(errno));
    return 0;
  }
  for(hport = conn->server->minPasvPort; hport <= conn->server->maxPasvPort;
      hport++)
  {
    *port = htons(hport);
    
    if(!bind(conn->dataSock, (struct sockaddr*)&conn->lDataAddr, al))
      break;
  }
  
  if(hport > conn->server->maxPasvPort)
  {
    syslog (LOG_ERR, "%d: Failed to bind passive socket: %m", conn->sock);
    reply(conn, "425 Failed to bind socket: %s.", strerror(errno));
    close(conn->dataSock);
    conn->dataSock = -1;
    return 0;
  }
  
  if(listen(conn->dataSock, 1))
  {
    syslog (LOG_ERR, "%d: Failed to listen on passive socket: %m", conn->sock);
    reply(conn, "425 Failed to listen on socket: %s.", strerror(errno));
    close(conn->dataSock);
    conn->dataSock = -1;
    return 0;
  }
  
  conn->passive = 1;
  add_read_fd(conn->dataSock, passive_acceptor, conn);
  
  reply(conn, "229 Entering extended passive mode (|||%d|)", hport);
  return 0;
}

int command_mode(connection_t *conn, const char *arg, int expected)
{
  switch(arg[0])
  {
  case 0:
    reply (conn, "501 Argument missing.");
    return 0;
  case 'S':
  case 's':
    reply (conn, "200 Mode set to Stream.");
    return 0;
  case 'B':
  case 'b':
    reply (conn, "504 Mode Block unsupported.");
    return 0;
  case 'C':
  case 'c':
    reply (conn, "504 Mode Compressed unsupported.");
    return 0;
  }
  
  reply (conn, "501 Unrecognised mode.");
  return 0;
}

int command_type(connection_t *conn, const char *arg, int expected)
{
  switch(arg[0])
  {
  case 0:
    reply (conn, "501 Argument missing.");
    return 0;
  case 'A':
  case 'a':
    if (arg[1] == ' ')
    {
      switch (arg[2])
      {
      case 0:
      case 'N':
	conn->subtype = 0;
	reply (conn, "200 Type set to Non-Print Text.");
	break;
      case 'T':
	conn->subtype = 'T';
	reply (conn, "504 Type Telnet Formatted Text unsupported.");
	return 0;
      case 'C':
	reply (conn, "504 Type Carriage Control Text unsupported.");
	return 0;
      }
    }
    else
    {
      conn->subtype = 0;
      reply(conn, "200 Type set to Non-Print Text.");
    }
    conn->type = ttText;
    return 0;
  case 'E':
  case 'e':
    reply (conn, "504 Type EBCDIC unsupported.");
    return 0;
  case 'I':
  case 'i':
    reply (conn, "200 Type set to Image.");
    conn->type = ttImage;
    conn->subtype = 0;
    return 0;
  case 'L':
  case 'l':
    if (arg[1] == ' ')
    {
      if (atoi (arg + 2) == 8)
      {
	conn->type = ttImage;
	conn->subtype = 8;
	reply (conn, "200 Type set to Local 8.");
      }
      else
	reply (conn, "504 Unsupported local size.");
    }
    else
      reply (conn, "501 Missing argument to L");
    return 0;
  }
  
  reply(conn, "501 Unrecognised type.");
  return 0;
}

int command_stru(connection_t *conn, const char *arg, int expected)
{
  switch(arg[0])
  {
  case 0:
    reply (conn, "501 Argument missing.");
    return 0;
  case 'F':
  case 'f':
    reply (conn, "200 File structure set to File.");
    return 0;
  case 'R':
  case 'r':
    reply (conn, "504 File structure Record unsupported.");
    return 0;
  case 'P':
  case 'p':
    reply (conn, "504 File structure Page unsupported.");
    return 0;
  }
  
  reply(conn, "501 Unrecognised file structure.");
  return 0;
}

/* File commands */

int command_allo(connection_t *conn, const char *arg, int expected)
{
  reply(conn, "202 ALLO not needed here.");
  return 0;
}

int command_rest(connection_t *conn, const char *arg, int expected)
{
  conn->restart = strtoull(arg, NULL, 10);
  reply(conn, "350 Continuing at %llu. Expecting transfer command.",
	conn->restart);
  conn->expect = "STOR|RETR";
  return 0;
}

int command_stor(connection_t *conn, const char *arg, int expected)
{
  struct stat st;
  
  if (!strlen (arg))
  {
    reply (conn, "501 Path missing.");
    return 0;
  }
  
  if(!expected)
    conn->restart = 0;
  
  if(conn->epsvOnly && !conn->passive)
  {
    reply(conn, "503 STOR without preceeding EPSV.");
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  
  arg = chroot_path(arg, conn->user->home);
  if(!check_access(arg, acSearch) && !lstat(arg, &st))
  {
    if(S_ISDIR(st.st_mode))
    {
      reply(conn, "553 Is a directory.");
      if(close_data_connection (conn) == -1)
	return 1;
      return 0;
    }
    if(S_ISLNK(st.st_mode))
    {
      reply(conn, "553 Writing to symlinks not allowed.");
      if(close_data_connection (conn) == -1)
	return 1;
      return 0;
    }
    if(S_ISBLK(st.st_mode))
    {
      reply(conn, "553 Writing to block devices not allowed.");
      if(close_data_connection (conn) == -1)
	return 1;
      return 0;
    }
    /* We do allow writes to character devices, sockets and FIFOs */
    if (conn->restart && conn->type == ttText)
    {
      long long l = text_type_restart (arg, conn->restart);
      if (l == -1)
      {
	reply (conn, "550 %s.", strerror (errno));
	if (close_data_connection (conn) == -1)
	  return 1;
	return 0;
      }
      conn->restart = l;
    }
    if(conn->restart > st.st_size)
    {
      reply(conn, "500 File smaller than restart offset.");
      if(close_data_connection (conn) == -1)
	return 1;
      return 0;
    }
  }
  else switch(errno)
  {
  case ENOENT:
    if(conn->restart)
    {
      reply(conn, "500 Can't restart a nonexistant file.");
      if(close_data_connection (conn) == -1)
	return 1;
      return 0;
    }
    break;
  case EBUSY:
    reply(conn, "450 File busy.");
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  default:
    reply(conn, "553 %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  
  if (check_prot_level (conn, arg))
    return 0;
  
  conn->fileFd = open_file_writing(arg, conn->restart, 0);
  if(conn->fileFd < 0)
  {
    if(errno == EBUSY)
      reply(conn, "450 File busy.");
    else
      reply(conn, "553 Failed to open file: %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  
  conn->sending = 0;
  if(open_data_connection(conn))
  {
    reply(conn, "425 Failed to open data connection: %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  conn->filePath = pstring (arg, conn);
  reply(conn, "150 Opening %s data connection for retrieval of \"%s\".",
	conn->type == ttImage? "BINARY" : "ASCII", print_path(arg));
  conn->working = 1;
  return 0;
}

int command_stou(connection_t *conn, const char *arg, int expected)
{
  const char template[] = "moftpd-upload.XXXXXX";
  char *path;
  
  if(conn->epsvOnly && !conn->passive)
  {
    reply(conn, "503 STOU without preceeding EPSV.");
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  
  path = chroot_path(template, NULL);
  
  if (check_prot_level (conn, path))
    return 0;
  
  conn->fileFd = open_temp_file(path);
  if(conn->fileFd < 0)
  {
    reply(conn, "553 Failed to open file: %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  
  conn->sending = 0;
  if(open_data_connection(conn))
  {
    reply(conn, "425 Failed to open data connection: %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  conn->filePath = pstring (path, conn);
  // Ignoring RFC 1123 for now, as I think it's important to relay the transfer type.
  reply(conn, "150 Opening %s data connection for retrieval of \"%s\".",
	conn->type == ttImage? "BINARY" : "ASCII", print_path(path));
  conn->working = 1;
  return 0;
}

int command_retr(connection_t *conn, const char *arg, int expected)
{
  struct stat st;
  
  if (!strlen (arg))
  {
    reply (conn, "501 Path missing.");
    return 0;
  }
  
  if(!expected)
    conn->restart = 0;
  
  if(conn->epsvOnly && !conn->passive)
  {
    reply(conn, "503 RETR without preceeding EPSV.");
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  
  arg = chroot_path(arg, conn->user->home);
  if(!check_access(arg, acSearch) && !stat(arg, &st))
  {
    if(S_ISDIR(st.st_mode))
    {
      reply(conn, "553 Is a directory.");
      if(close_data_connection (conn) == -1)
	return 1;
      return 0;
    }
    if (conn->restart && conn->type == ttText)
    {
      long long l = text_type_restart (arg, conn->restart);
      if (l == -1)
      {
	reply (conn, "550 %s.", strerror (errno));
	if (close_data_connection (conn) == -1)
	  return 1;
	return 0;
      }
      conn->restart = l;
    }
    if(conn->restart > st.st_size)
    {
      reply(conn, "500 File smaller than restart offset.");
      if(close_data_connection (conn) == -1)
	return 1;
      return 0;
    }
  }
  else
  {
    reply(conn, "550 %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  
  if (check_prot_level (conn, arg))
    return 0;
  
  conn->fileFd = open_file_reading(arg, conn->restart);
  if(conn->fileFd < 0)
  {
    if(errno == EBUSY)
      reply(conn, "450 File busy.");
    else
      reply(conn, "553 Failed to open file: %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
#ifdef HAVE_MMAP
  if(conn->type == ttImage && st.st_size <= maxMmapSize)
    conn->data = mmap(NULL, st.st_size, PROT_READ, MAP_FILE | MAP_PRIVATE,
	  conn->fileFd, 0);
  else
#endif
    conn->data = (char*)-1;
  if((int)conn->data != -1)
  {
    conn->dataLen = st.st_size;
    conn->dataOffset = conn->restart;
  }
  else
    conn->data = NULL;
  
  conn->sending = 1;
  if(open_data_connection(conn))
  {
    reply(conn, "425 Failed to open data connection: %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  conn->filePath = pstring (arg, conn);
  reply(conn, "150 Opening %s data connection for sending of \"%s\" (%lld bytes).",
	conn->type == ttImage? "BINARY" : "ASCII", print_path(arg), (long long)st.st_size);
  conn->working = 1;
  return 0;
}

int command_list(connection_t *conn, const char *arg, int expected)
{
  if(conn->epsvOnly && !conn->passive)
  {
    reply(conn, "503 LIST without preceeding EPSV.");
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  
  if(!strlen(arg))
    arg = ".";
  
  conn->data = pstring(make_file_list(arg, conn->user->home, 1, NULL), conn);
  if(!conn->data)
  {
    reply(conn, "550 Failed to list \"%s\": %s.", arg, strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  conn->dataOffset = 0;
  conn->dataLen = strlen(conn->data);
  
  conn->sending = 1;
  if(open_data_connection(conn))
  {
    reply(conn, "425 Failed to open data connection: %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  reply(conn, "150 Opening data connection for listing of \"%s\".",
	print_path(arg));
  conn->working = 1;
  return 0;
}

int command_nlst(connection_t *conn, const char *arg, int expected)
{
  if(conn->epsvOnly && !conn->passive)
  {
    reply(conn, "503 NLST without preceeding EPSV.");
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  
  if(!strlen(arg))
    arg = ".";
  
  conn->data = pstring (make_file_list(arg, conn->user->home, 0, NULL), conn);
  if(!conn->data)
  {
    reply(conn, "550 Failed to list \"%s\": %s.", arg, strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  conn->dataOffset = 0;
  conn->dataLen = strlen(conn->data);
  
  conn->sending = 1;
  if(open_data_connection(conn))
  {
    reply(conn, "425 Failed to open data connection: %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  reply(conn, "150 Opening data connection for listing of \"%s\".",
	print_path(arg));
  conn->working = 1;
  return 0;
}

int command_mlst(connection_t *conn, const char *arg, int expected)
{
  const char *mlst = NULL;
  const char *path;
  
  if (!strlen (arg))
    arg = ".";
  
  path = chroot_path (arg, conn->user->home);
  
  if (!check_access (path, acSearch))
    mlst = tagged_file_data (path, print_path (path), conn->mlstTags);
  if (mlst)
  {
    reply (conn, "250-Data about \"%s\":", arg);
    reply (conn, " %s", mlst);
    reply (conn, "250 End of data.");
  }
  else
    reply (conn, "550 %s.", strerror (errno));
  return 0;
}

int command_mlsd(connection_t *conn, const char *arg, int expected)
{
  const char *path;
  
  if (conn->epsvOnly && !conn->passive)
  {
    reply (conn, "503 MLSD without preceeding EPSV.");
    if (close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  
  if (!strlen (arg))
    arg = ".";
  
  path = chroot_path (arg, conn->user->home);
  
  conn->data = pstring (tagged_file_list(path, conn->mlstTags), conn);
  if(!conn->data)
  {
    if (errno == ENOTDIR)
      reply (conn, "501 \"%s\" is not a directory.", arg);
    else
      reply (conn, "550 Failed to list \"%s\": %s.", arg, strerror (errno));
    if (close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  conn->dataOffset = 0;
  conn->dataLen = strlen(conn->data);
  
  conn->sending = 1;
  if(open_data_connection(conn))
  {
    reply(conn, "425 Failed to open data connection: %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  reply(conn, "150 Opening data connection for listing of \"%s\".",
	print_path(arg));
  conn->working = 1;
  return 0;
}

int command_appe(connection_t *conn, const char *arg, int expected)
{
  struct stat st;
  
  if (!strlen (arg))
  {
    reply (conn, "501 Path missing.");
    return 0;
  }
  
  if(!expected)
    conn->restart = 0;
  
  if(conn->epsvOnly && !conn->passive)
  {
    reply(conn, "503 APPE without preceeding EPSV.");
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  
  arg = chroot_path(arg, conn->user->home);
  if(!check_access(arg, acSearch) && !lstat(arg, &st))
  {
    if(S_ISDIR(st.st_mode))
    {
      reply(conn, "553 Is a directory.");
      if(close_data_connection (conn) == -1)
	return 1;
      return 0;
    }
    if(S_ISLNK(st.st_mode))
    {
      reply(conn, "553 Writing to symlinks not allowed.");
      if(close_data_connection (conn) == -1)
	return 1;
      return 0;
    }
    if(S_ISBLK(st.st_mode))
    {
      reply(conn, "553 Writing to block devices not allowed.");
      if(close_data_connection (conn) == -1)
	return 1;
      return 0;
    }
    /* We do allow writes to character devices, sockets and FIFOs */
  }
  else switch(errno)
  {
  case ENOENT:
    reply(conn, "550 File does not exist.");
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  default:
    reply(conn, "553 %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  
  if (check_prot_level (conn, arg))
    return 0;
  
  conn->fileFd = open_file_appending(arg);
  if(conn->fileFd < 0)
  {
    if(errno == EBUSY)
      reply(conn, "450 File busy.");
    else
      reply(conn, "553 Failed to open file: %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  if(lseek(conn->fileFd, st.st_size, SEEK_SET) == -1)
  {
    reply(conn, "550 Failed to seek in file: %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  
  conn->sending = 0;
  if(open_data_connection(conn))
  {
    reply(conn, "425 Failed to open data connection: %s.", strerror(errno));
    if(close_data_connection (conn) == -1)
      return 1;
    return 0;
  }
  conn->filePath = pstring (arg, conn);
  reply(conn, "150 Opening %s data connection for appending of \"%s\".",
	conn->type == ttImage? "BINARY" : "ASCII", print_path(arg));
  conn->working = 1;
  return 0;
}

int command_rnfr(connection_t *conn, const char *arg, int expected)
{
  struct stat st;
  
  if (!strlen (arg))
  {
    reply (conn, "501 Path missing.");
    return 0;
  }
  
  arg = chroot_path(arg, conn->user->home);
  if(check_access(arg, acRename) || lstat(arg, &st))
  {
    switch(errno)
    {
    case ENOENT:
      reply(conn, "550 File does not exist.");
      return 0;
    default:
      reply(conn, "553 %s.", strerror(errno));
      return 0;
    }
  }
  
  if(conn->rnfr)
    pfree(conn->rnfr, conn);
  conn->rnfr = pstring(arg, conn);
  if(!conn->rnfr)
  {
    reply(conn, "553 %s.", strerror(errno));
    return 0;
  }
  
  reply(conn, "350 RNTO expected.");
  conn->expect = "RNTO";
  return 0;
}

int command_rnto(connection_t *conn, const char *arg, int expected)
{
  struct stat st;
  
  if (!strlen (arg))
  {
    reply (conn, "501 Path missing.");
    return 0;
  }
  
  if(!expected)
  {
    reply(conn, "503 RNTO not expected here.");
    return 0;
  }
  
  arg = chroot_path(arg, conn->user->home);
  if(check_access(arg, acRename))
  {
    reply(conn, "553 %s.", strerror(errno));
    return 0;
  }
  
  if(!lstat(arg, &st))
  {
    reply(conn, "553 File exists.");
    return 0;
  }
  
  if(rename_file(conn->rnfr, arg))
  {
    pfree(conn->rnfr, conn);
    conn->rnfr = NULL;
    reply(conn, "553 %s.", strerror(errno));
    return 0;
  }
  pfree(conn->rnfr, conn);
  conn->rnfr = NULL;
  reply(conn, "250 Rename ok.");
  return 0;
}

int command_dele(connection_t *conn, const char *arg, int expected)
{
  if (!strlen (arg))
  {
    reply (conn, "501 Path missing.");
    return 0;
  }
  
  arg = chroot_path(arg, conn->user->home);
  if(delete_file(arg))
  {
    switch(errno)
    {
    case ENOENT:
      reply(conn, "550 File does not exist.");
      return 0;
    default:
      reply(conn, "553 %s.", strerror(errno));
      return 0;
    }
  }
  
  reply(conn, "250 File deleted.");
  return 0;
}

int command_rmd(connection_t *conn, const char *arg, int expected)
{
  if (!strlen (arg))
  {
    reply (conn, "501 Path missing.");
    return 0;
  }
  
  arg = chroot_path(arg, conn->user->home);
  if(delete_dir(arg))
  {
    switch(errno)
    {
    case ENOENT:
      reply(conn, "550 File does not exist.");
      return 0;
    default:
      reply(conn, "553 %s.", strerror(errno));
      return 0;
    }
  }
  
  reply(conn, "250 Directory deleted.");
  return 0;
}

int command_mkd(connection_t *conn, const char *arg, int expected)
{
  if (!strlen (arg))
  {
    reply (conn, "501 Path missing.");
    return 0;
  }
  
  arg = chroot_path(arg, conn->user->home);
  if(create_dir(arg))
  {
    reply(conn, "553 %s.", strerror(errno));
    return 0;
  }
  
  reply(conn, "250 Directory created.");
  return 0;
}

int command_pwd(connection_t *conn, const char *arg, int expected)
{
  reply(conn, "257 \"%s\" is current directory.", print_path(chroot_path(".",
	NULL)));
  return 0;
}

int command_abor(connection_t *conn, const char *arg, int expected)
{
  switch (close_data_connection (conn))
  {
  case -1:
    return 1;
  case 0:
    reply (conn, "226 Abort successful.");
    return 0;
  default:
    reply (conn, "426 Data connection closed.");
    reply (conn, "226 Abort successful.");
    return 0;
  }
}

/* Informational commands */

int command_syst(connection_t *conn, const char *arg, int expected)
{
  reply(conn, "215 UNIX system type.");
  return 0;
}

int command_stat(connection_t *conn, const char *arg, int expected)
{
  if(conn->working)
  {
    if(conn->data)
      reply(conn, "213 Sent %llu of %llu bytes.",
	    conn->dataOffset - conn->restart, conn->dataLen - conn->restart);
    else if(conn->sending)
      reply(conn, "213 Sent %llu bytes.", conn->dataOffset);
    else
      reply(conn, "213 Received %llu bytes.", conn->dataOffset);
  }
  else
  {
    if(strlen(arg))
    {
      char *list = make_file_list(arg, conn->user->home, 1, " ");
      if(!list)
      {
	if(errno == ENOENT)
	{
	  reply(conn, "500 No such file or directory.");
	  return 0;
	}
	reply(conn, "450 %s.", strerror(errno));
	return 0;
      }
      reply (conn, "212-Listing of \"%s\":", arg);
      send_telnet (conn, list, 0);
      reply (conn, "212 End of listing.");
    }
    else
    {
      reply (conn, "211-Connection status:");
      reply (conn, " Transfer type is %s.", conn->type == ttImage ? 
	    conn->subtype? "Local 8" : "Image" :
	    conn->subtype? "Telnet Formatted Text" : "Non-Print Text");
      reply (conn, " Transfer mode is Stream.");
      reply (conn, " File structure is File.");
      reply (conn, " Data connection is %s.", conn->dataSock >= 0 ?
	    !conn->passive || conn->passiveAccepted ? "open" :
	    "waiting for connect" : "closed");
      reply (conn, "211 End of status.");
    }
  }
  return 0;
}

int command_size(connection_t *conn, const char *arg, int expected)
{
  struct stat st;
  long long res;
  
  if (!strlen (arg))
  {
    reply (conn, "501 Path missing.");
    return 0;
  }
  
  arg = chroot_path(arg, conn->user->home);
  if (!check_access (arg, acReadFile) && !stat (arg, &st))
  {
    if(S_ISDIR(st.st_mode))
    {
      reply(conn, "553 Is a directory.");
      return 0;
    }
  }
  else
  {
    reply(conn, "550 %s.", strerror(errno));
    return 0;
  }
  
  res = st.st_size;
  if (conn->type == ttText)
  {
    int fd = open_file_reading (arg, 0);
    int l;
    char buf[1024], *bp;
    
    if (fd < 0)
    {
      reply (conn, "550 %s.", strerror (errno));
      return 0;
    }
    
    while ((l = read (fd, buf, 1024)) > 0)
    {
      bp = buf;
      while (l > 0 && (bp = memchr (bp, '\n', l)))
      {
	res++;
	l -= ++bp - buf;
      }
    }
    
    close (fd);
  }
  reply(conn, "213 %lld", res);
  return 0;
}

int command_mdtm(connection_t *conn, const char *arg, int expected)
{
  struct stat st;
  
  if (!strlen (arg))
  {
    reply (conn, "501 Path missing.");
    return 0;
  }
  
  arg = chroot_path (arg, conn->user->home);
  if (!check_access (arg, acReadFile) && !stat (arg, &st))
  {
    struct tm *tm = gmtime (&st.st_mtime);
    
    // TODO: fractions
    reply (conn, "213 %04d%02d%02d%02d%02d%02d", tm->tm_year + 1900, tm->tm_mon,
	  tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
  }
  else
    reply (conn, "550 %s.", strerror (errno));
  return 0;
}

int command_help(connection_t *conn, const char *arg, int expected)
{
  const int numCommands = sizeof(commands) / sizeof(command_t) - 1;
  /* gcc says spbuf is variable-sized if I just const int. */
#define cwidth 6
  const int columns = 80 / cwidth;
  const int rows = (numCommands + columns - 1) / columns;
  const command_t *comm[columns];
  int i, j;
  char line[81], *lend;
  
  reply(conn, "214-These commands are recognised:");
  for(i = 0; i < columns; i++)
  {
    if (rows * i < numCommands)
      comm[i] = &commands[rows * i];
    else
      comm[i] = &commands[numCommands];
  }
  while(comm[0] != &commands[rows])
  {
    strcpy (line, " ");
    lend = line + 1;
    for(i = 0; i < columns; i++)
    {
      if(comm[i]->name)
      {
	strcpy (lend, comm[i]->name);
	j = strlen (lend);
	lend += j;
	for (; j < cwidth; j++)
	  *lend++ = ' ';
	comm[i]++;
      }
      else
	break;
    }
    *lend = 0;
    reply (conn, "%s", line);
  }
  
  reply(conn, "214 End of help.");
  return 0;
#undef cwidth
}

int command_feat (connection_t *conn, const char *arg, int expected)
{
  DIR *locDir;
  struct dirent *dp;
  char *ltxt, *suff, *lp, *np;
  
  reply (conn, "211-These features are supported.");
  //reply (conn, " ADAT");
  reply (conn, " AUTH"
#ifdef HAVE_LIBSSL
	" TLS"
#endif
	 );
  //reply (conn, " CCC");
  //reply (conn, " CONF");
  //reply (conn, " ENC");
  reply (conn, " EPRT");
  reply (conn, " EPSV");
  if ((locDir = opendir (localeDir)))
  {
    ltxt = talloc (4);
    if (ltxt)
    {
      strcpy (ltxt, "EN");
      lp = ltxt + 2;
      if (!conn->currLang)
	*lp++ = '*';
      while ((dp = readdir (locDir)))
      {
	if (dp->d_name[0] == '.')
	  continue;
	suff = strstr (dp->d_name, localeSuffix);
	if (!suff || strlen (suff) != strlen (localeSuffix))
	  continue;
	ltxt = trealloc (ltxt, lp - ltxt + suff - dp->d_name + 3);
	if (!ltxt)
	  break;
	lp = ltxt + strlen (ltxt);
	*lp++ = ';';
	for (np = dp->d_name; np != suff; np++)
	  *lp++ = toupper (*np);
	if (conn->currLang && !strncmp (dp->d_name, conn->currLang, suff - dp->d_name))
	  *lp++ = '*';
      }
      if (ltxt)
      {
	*lp = 0;
	reply (conn, " LANG %s", ltxt);
      }
    }
    closedir (locDir);
  }
  else
    reply (conn, " LANG EN*");
  reply (conn, " LPRT");
  reply (conn, " LPSV");
  reply (conn, " MDTM");
  //reply (conn, " MIC");
  reply (conn, " MLST TYPE%s;UNIQUE%s;MODIFY%s;PERM%s;SIZE%s;",
	conn->mlstTags & tfType? "*" : "", conn->mlstTags & tfUnique? "*" : "",
	conn->mlstTags & tfModify? "*" : "", conn->mlstTags & tfPerm? "*" : "",
	conn->mlstTags & tfSize? "*" : "");
  reply (conn, " PBSZ");
  reply (conn, " PROT");
  reply (conn, " REST STREAM");
  reply (conn, " SIZE");
  reply (conn, " TVFS");
  reply (conn, " UTF8");
  reply (conn, "211 End of features list.");
  return 0;
}

/* Settings commands */

int command_opts (connection_t *conn, const char *arg, int expected)
{
  const char *ap;
  
  if (!strlen (arg))
  {
    reply (conn, "501 Argument missing.");
    return 0;
  }
  if (!strncasecmp (arg, "MLST", 4) && (arg[4] == ' ' || !arg[4]))
  {
    conn->mlstTags = 0;
    if (arg[4])
    {
      arg += 5;
      while ((ap = strchr (arg, ';')))
      {
	switch (ap - arg)
	{
	case 4:
	  if (!strncasecmp (arg, "TYPE", 4))
	    conn->mlstTags |= tfType;
	  else if (!strncasecmp (arg, "PERM", 4))
	    conn->mlstTags |= tfPerm;
	  else if (!strncasecmp (arg, "SIZE", 4))
	    conn->mlstTags |= tfSize;
	  break;
	case 6:
	  if (!strncasecmp (arg, "UNIQUE", 6))
	    conn->mlstTags |= tfUnique;
	  else if (!strncasecmp (arg, "MODIFY", 6))
	    conn->mlstTags |= tfModify;
	  break;
	}
	arg = ap + 1;
      }
    }
    reply (conn, "200 MLST OPTS %s%s%s%s%s",
	  conn->mlstTags & tfType? "TYPE;" : "",
	  conn->mlstTags & tfUnique? "UNIQUE;" : "",
	  conn->mlstTags & tfModify? "MODIFY;" : "",
	  conn->mlstTags & tfPerm? "PERM;" : "",
	  conn->mlstTags & tfSize? "SIZE;" : "");
    return 0;
  }
  reply (conn, "501 No such setting.");
  return 0;
}

int command_lang(connection_t *conn, const char *arg, int expected)
{
  char *tmp, *tp, *lp;
  int fd;
  
  if (!strlen (arg))
    arg = "en";
  
  tmp = talloc (strlen (localeDir) + strlen (arg) + strlen (localeSuffix) + 1);
  if (!tmp)
  {
    reply (conn, "500 %s", strerror (errno));
    return 0;
  }
  strcpy (tmp, localeDir);
  lp = tp = tmp + strlen (tmp);
  strcpy (tp, arg);
  for (; *tp; tp++)
    *tp = tolower (*tp);
  
  while (tp)
  {
    *tp = 0;
    if (!strcmp (lp, "en"))
    {
      if (conn->langFd >= 0)
	close_shared (conn->langFd);
      conn->langFd = -1;
      pfree (conn->currLang, conn);
      conn->currLang = NULL;
      reply (conn, "200 Language set to English.");
      return 0;
    }
    strcpy (tp, localeSuffix);
    fd = open_shared (tmp);
    if (fd >= 0)
    {
      if (conn->langFd >= 0)
	close_shared (conn->langFd);
      conn->langFd = fd;
      pfree (conn->currLang, conn);
      *tp = 0;
      conn->currLang = pstring (lp, conn);
      reply (conn, "200 Language set to English.");
      return 0;
    }
    *tp = 0;
    tp = strrchr (lp, '-');
  }
  reply (conn, "504 Unknown language.");
  return 0;
}

/* Miscellaneous commands */

int command_site(connection_t *conn, const char *arg, int expected)
{
  reply(conn, "502 SITE not yet implemented.");
  return 0;
}

int command_noop(connection_t *conn, const char *arg, int expected)
{
  reply(conn, "200 No action taken.");
  return 0;
}

/* Security commands */

int command_auth (connection_t *conn, const char *arg, int expected)
{
  if (!strlen (arg))
  {
    reply (conn, "501 Argument missing.");
    return 0;
  }
  if (conn->authed)
  {
    reply (conn, "534 Can't start encryption when logged in.");
    return 0;
  }
#ifdef HAVE_LIBSSL
  if (conn->ssl)
  {
    reply (conn, "534 Authentication already negotiated.");
    return 0;
  }
  
  if (!strcasecmp (arg, "TLS") || !strcasecmp (arg, "TLC-C"))
  {
    BIO *bio;
    SSL *ssl;
    
    if (!conn->server->tlsCert)
    {
      reply (conn, "534 This server does not allow encryption.");
      return 0;
    }
    if (sslCtx == (SSL_CTX *)-1)
    {
      reply (conn, "431 TLS initialisation failed.");
      return 0;
    }
    
    if (!sslCtx)
    {
      SSL_load_error_strings ();
      SSL_library_init ();
      
      sslCtx = SSL_CTX_new (SSLv23_server_method ());
      if (!sslCtx)
      {
	reply (conn, "431 TLS initialisation failed.");
	// We only try once.
	sslCtx = (SSL_CTX*)-1;
	return 0;
      }
      SSL_CTX_load_verify_locations (sslCtx, NULL, sslCertsPath);
    }
    
    ssl = SSL_new (sslCtx);
    if (!ssl)
    {
      reply (conn, "431 TLS initialisation failed.");
      return 0;
    }
    SSL_set_options (ssl, conn->server->tlsOptions);
    if (!SSL_use_certificate (ssl, conn->server->tlsCert))
    {
      syslog (LOG_ERR, "Failed to load certificate file: %s",
	    ERR_reason_error_string (ERR_get_error ()));
      reply (conn, "431 TLS initialisation failed.");
      SSL_free (ssl);
      return 0;
    }
    if (!SSL_use_PrivateKey (ssl, conn->server->tlsKey))
    {
      syslog (LOG_ERR, "Failed to load private key: %s",
	    ERR_reason_error_string (ERR_get_error ()));
      reply (conn, "431 TLS initialisation failed.");
      SSL_free (ssl);
      return 0;
    }
    if (conn->server->verifyTLS)
    {
      STACK_OF(X509_NAME) *caStack;
      
      SSL_set_verify (ssl, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
	    NULL);
      
      caStack = sk_X509_new_null ();
      SSL_add_dir_cert_subjects_to_stack (caStack, sslCertsPath);
      SSL_set_client_CA_list (ssl, caStack);
    }
    bio = BIO_new_socket (conn->sock, BIO_NOCLOSE);
    if (!bio)
    {
      SSL_free (ssl);
      reply (conn, "431 TLS initialisation failed.");
      return 0;
    }
    BIO_set_nbio (bio, 1);
    reply (conn, "234 Waiting for TLS handshake.");
    SSL_set_bio (ssl, bio, bio);
    conn->ssl = ssl;
    conn->sslState = 0;
    conn->pbsz = -1;
    return 0;
  }
#endif
  reply (conn, "504 Unknown authentication method.");
  return 0;
}

int command_adat (connection_t *conn, const char *arg, int expected)
{
  reply (conn, "502 ADAT not yet implemented.");
  return 0;
}

int command_prot (connection_t *conn, const char *arg, int expected)
{
  int lvl;
  
  if (!strlen (arg))
  {
    reply (conn, "501 Argument missing.");
    return 0;
  }
  
  if (conn->dataSock >= 0)
  {
    reply (conn, "503 Can't change protection level with open data socket.");
    return 0;
  }
  
#ifdef HAVE_LIBSSL
  if (!conn->ssl)
#endif
  {
    reply (conn, "536 Need to establish a secure connection first.");
    return 0;
  }
  
  if (conn->pbsz == -1)
  {
    reply (conn, "503 Need to negotiate buffer size first.");
    return 0;
  }
  
  switch (arg[0])
  {
  case 'C':
  case 'c':
    lvl = 0;
    break;
  case 'S':
  case 's':
  case 'E':
  case 'e':
    reply (conn, "536 TLS does not support this protection level.");
    return 0;
  case 'P':
  case 'p':
    lvl = acEncrypted | acSigned;
    break;
  default:
    reply (conn, "504 Unknown protection level.");
    return 0;
  }
  conn->protLevel = lvl;
  reply (conn, "200 Protection level set.");
  return 0;
}

int command_pbsz (connection_t *conn, const char *arg, int expected)
{
  char *ep;
  /*unsigned long l = */strtoul (arg, &ep, 10);
  
  if (!strlen (arg))
  {
    reply (conn, "501 Argument missing.");
    return 0;
  }
  if (ep && *ep)
  {
    reply (conn, "501 Numerical argument expected.");
    return 0;
  }
  
#ifdef HAVE_LIBSSL
  if (conn->ssl)
  {
    conn->pbsz = 0;
    reply (conn, "200 PBSZ=0");
  }
  else
#endif
    reply (conn, "503 Need to establish a secure connection first.");
  return 0;
}

int command_ccc (connection_t *conn, const char *arg, int expected)
{
#ifdef HAVE_LIBSSL
  if (conn->ssl)
  {
    reply (conn, "534 Won't turn off TLS protection.");
    return 0;
  }
  else
#endif
    reply (conn, "503 Need to establish a secure connection first.");
  return 0;
}

int command_mic (connection_t *conn, const char *arg, int expected)
{
  reply (conn, "502 MIC not yet implemented.");
  return 0;
}

int command_conf (connection_t *conn, const char *arg, int expected)
{
  reply (conn, "502 CONF not yet implemented.");
  return 0;
}

int command_enc (connection_t *conn, const char *arg, int expected)
{
  reply (conn, "502 ENC not yet implemented.");
  return 0;
}
