/*****************************************************************************\
* Copyright (c) 2002 Pelle Johansson.                                         *
* All rights reserved.                                                        *
*                                                                             *
* This file is part of the moftpd package. Use and distribution of            *
* this software is governed by the terms in the file LICENCE, which           *
* should have come with this package.                                         *
\*****************************************************************************/

/* $moftpd: confparse.c,v 1.16.2.4 2003/09/29 12:58:31 morth Exp $ */
/* $Id$ */

#include "confparse.h"

#include "config.tab.h"
#include "memory.h"

#include "system.h"

static int confFd;
static char *confBuf, *confP;
static int confLine;

extern int confExpectingToken;

typedef struct identifier
{
  const char *name;
  int rule;
} identifier_t;

static identifier_t identifiers[] =
{
  {"Server"          , I_SERVER          },
  {"User"            , I_USER            },
  {"Directory"       , I_DIRECTORY       },
  
  {"UnprivUid"       , I_UNPRIVUID       },
  {"UnprivGid"       , I_UNPRIVGID       },
  {"XferBuffSize"    , I_XFERBUFFSIZE    },
  {"LocaleDir"       , I_LOCALEDIR       },
  {"FakeChroot"      , I_FAKECHROOT      },
  {"ForkClients"     , I_FORKCLIENTS     },
  {"Reset"           , I_RESET           },
  {"exec"            , I_EXEC            },
  {"TrustedCertsDir" , I_TRUSTEDCERTSDIR },
  {"MaxMmapSize"     , I_MAXMMAPSIZE     },
  
  {"AllowForeign"    , I_ALLOWFOREIGN    },
  {"AllowLowPorts"   , I_ALLOWLOWPORTS   },
  {"AllowUnbound"    , I_ALLOWUNBOUND    },
  {"Bind"            , I_BIND            },
  {"MinPasvPort"     , I_MINPASVPORT     },
  {"MaxPasvPort"     , I_MAXPASVPORT     },
  {"Port"            , I_PORT            },
  {"PassIfInvalid"   , I_PASSIFINVALID   },
  {"LoginFailedMsg"  , I_LOGINFAILEDMSG  },
  {"UserInvalidMsg"  , I_USERINVALIDMSG  },
  {"PassRequestMsg"  , I_PASSREQUESTMSG  },
  {"AnonPassMsg"     , I_ANONPASSMSG     },
  {"AcceptTLS"       , I_ACCEPTTLS       },
  {"TLSVerifyClient" , I_TLSVERIFYCLIENT },
  {"TLSAutoLogin"    , I_TLSAUTOLOGIN    },
  {"TLSNoNewUser"    , I_TLSNONEWUSER    },
  {"UserAlias"       , I_USERALIAS       },
  {"SleepOnFail"     , I_SLEEPONFAIL     },
  {"MaxLoginAttempts", I_MAXLOGINATTEMPTS},
  
  {"Alias"           , I_ALIAS           },
  {"AllowLogin"      , I_ALLOWLOGIN      },
  {"AllowSecLogin"   , I_ALLOWSECLOGIN   },
  {"Anonymous"       , I_ANONYMOUS       },
  {"Chroot"          , I_CHROOT          },
  {"Gid"             , I_GID             },
  {"Home"            , I_HOME            },
  {"Password"        , I_PASSWORD        },
  {"PasswordNeeded"  , I_PASSWORDNEEDED  },
  {"Uid"             , I_UID             },
  {"MaxIdle"         , I_MAXIDLE         },
  
  {"Allow"           , I_ALLOW           },
  {"Deny"            , I_DENY            },
  {"Require"         , I_REQUIRE         },
  {"Hidden"          , I_HIDDEN          },
  {"FakeUser"        , I_FAKEUSER        },
  {"FakeGroup"       , I_FAKEGROUP       },
  {"FakeFileMode"    , I_FAKEFILEMODE    },
  {"FakeDirMode"     , I_FAKEDIRMODE     },
  
  {"search"          , I_SEARCH          },
  {"readFile"        , I_READFILE        },
  {"listing"         , I_LISTING         },
  {"createFile"      , I_CREATEFILE      },
  {"createDir"       , I_CREATEDIR       },
  {"append"          , I_APPEND          },
  {"overwrite"       , I_OVERWRITE       },
  {"delete"          , I_DELETE          },
  {"rename"          , I_RENAME          },
  {"encrypted"       , I_ENCRYPTED       },
  {"signed"          , I_SIGNED          },
  {"reading"         , I_READING         },
  {"writing"         , I_WRITING         },
  {"storing"         , I_STORING         },
  {"all"             , I_ALL             },
  
  {NULL}
};

int read_config(const char *path)
{
  confFd = open(path, O_RDONLY);
  if(confFd < 0)
  {
    syslog (LOG_ERR, "Failed to open config file: %m.");
    return -1;
  }
  confLine = 1;
  
  return yyparse();
}

static int fetch_ch(void)
{
  int l;
  
  if(confFd < 0)
    return 0;
  
  if (!confBuf)
  {
    confBuf = talloc (4097);
    if (!confBuf)
      return -1;
  }
  
  while(!confP || !*confP)
  {
    l = read(confFd, confBuf, 4096);
    if(l <= 0)
    {
      close(confFd);
      confFd = -1;
      return l;
    }
    
    confBuf[l] = 0;
    confP = confBuf;
  }
  
  if(*confP == '\n')
    confLine++;
  
  return *confP++;
}

static void unfetch_ch(void)
{
  if(confP && confP > confBuf)
  {
    confP--;
    if(*confP == '\n')
      confLine--;
  }
}

int yylex(void)
{
  int ch = fetch_ch(), sch;
  char strbuf[100], *sp;
  static identifier_t *ident;
  int expTok;
  
  expTok = confExpectingToken;
  confExpectingToken = 0;
  
  if(ch <= 0)
    return ch;
  
  if(isspace(ch) || ch == '#')
  {
    do
    {
      while(isspace(ch))
	ch = fetch_ch();
      if(ch == '#')
      {
	while((ch = fetch_ch()) != '\n')
	  if(ch <= 0)
	    return ch;
      }
    } while(isspace(ch));
    if(ch <= 0)
      return ch;
  }
  
  if(ch == '>' || ch == ';' || ch == ',')
    return ch;
  if (!expTok && ch == '<')
  {
    ch = fetch_ch ();
    if (ch == '/')
      return I_ETAG;
    unfetch_ch ();
    return '<';
  }
  
  if(ch == '"' || ch == '\'')
  {
    sch = ch;
    sp = strbuf;
    while(sp - strbuf < 100)
    {
      do
      {
	ch = fetch_ch();
	if(ch <= 0)
	  return -1;
	*sp++ = ch;
      } while(ch != sch && sp - strbuf < 100);
      
      ch = fetch_ch();
      if(ch != sch)
      {
	while(isspace(ch))
	  ch = fetch_ch();
	if(ch == sch)
	  sp--;
	else
	{
	  unfetch_ch();
	  break;
	}
      }
    }
    
    *(sp - 1) = 0;
    strcpy(yylval.str, strbuf);
    
    return D_STRING;
  }
  
  sp = strbuf;
  while(!isspace(ch) && ch != ',' && ch != '>' && ch != ';' && sp - strbuf <
	100)
  {
    *sp++ = ch;
    ch = fetch_ch();
    if(ch < 0)
      return -1;
    if(!ch)
      break;
  }
  
  *sp = 0;
  if(ch)
    unfetch_ch();
  
  switch (expTok)
  {
  case D_STRING:
    strcpy(yylval.str, strbuf);
    return D_STRING;
  case D_UID:
    yylval.num = strtoll(strbuf, &sp, 0);
    if (sp && *sp)
    {
      struct passwd *pwd;
      
      pwd = getpwnam (strbuf);
      if (!pwd)
	return -1;
      yylval.num = pwd->pw_uid;
    }
    return D_UID;
  case D_GID:
    yylval.num = strtoll(strbuf, &sp, 0);
    if (sp && *sp)
    {
      struct group *gr;
      
      gr = getgrnam (strbuf);
      if (!gr)
	return -1;
      yylval.num = gr->gr_gid;
    }
    return D_GID;
  }
  
  if(!strlen(strbuf))
    return -1;
  
  for(ident = identifiers; ident->name; ident++)
  {
    if(!strcasecmp(ident->name, strbuf))
      return ident->rule;
  }
  
  if(!strcasecmp(strbuf, "true") || !strcasecmp(strbuf, "on") ||
	!strcasecmp(strbuf, "yes"))
  {
    yylval.num = 1;
    return D_BOOL;
  }
  if(!strcasecmp(strbuf, "false") || !strcasecmp(strbuf, "off") ||
	!strcasecmp(strbuf, "no"))
  {
    yylval.num = 0;
    return D_BOOL;
  }
  
  yylval.num = strtoll(strbuf, &sp, 0);
  if(sp == strbuf)
    return -1;
  while(sp && *sp)
  {
    switch(*sp++)
    {
    case 'K':
      yylval.num *= 1024;
      break;
    case 'M':
      yylval.num *= 1024 * 1024;
      break;
    case 'G':
      yylval.num *= 1024 * 1024 * 1024;
      break;
    case 'm':
      yylval.num *= 60;
      break;
    case 'h':
      yylval.num *= 60 * 60;
      break;
    case 'd':
      yylval.num *= 60 * 60 * 24;
      break;
    default:
      return -1;
    }
  }
  return D_NUMBER;
}

void yyerror(const char *err)
{
  char *eol;
  
  // Trick to make sure there's data loaded if available.
  fetch_ch();
  unfetch_ch();
  
  eol = strchr(confP, '\n');
  if(eol)
    *eol = 0;
  
  syslog (LOG_ERR, "Config file line %d: %s near %.12s.", confLine,
	err, confP? eol == confP? "End-of-Line": confP : "End-of-File");
  
  if(eol)
    *eol = '\n';
}
