/*****************************************************************************\
* Copyright (c) 2002-2003 Pelle Johansson.                                    *
* All rights reserved.                                                        *
*                                                                             *
* This file is part of the moftpd package. Use and distribution of            *
* this software is governed by the terms in the file LICENCE, which           *
* should have come with this package.                                         *
\*****************************************************************************/

/* $moftpd: connection.c,v 1.37.2.8 2004/03/16 20:07:27 morth Exp $ */
/* $Id$ */

#include "connection.h"

#include "server.h"
#include "main.h"
#include "commands.h"
#include "utf8fs/file.h"
#include "memory.h"

#include "system.h"

extern int forkConnections, urgData;

int xferBufferSize = 1024;
char *xferBuffer;

#ifdef HAVE_LIBSSL
extern const char *sslCertsPath;
extern SSL_CTX *sslCtx;
#endif

static void *connections;

connection_t *new_connection(int sock, server_t *serv)
{
  connection_t *res;
  struct sockaddr_storage lAddr, rAddr;
  int lAl, rAl;
  char rhost[NI_MAXHOST], lhost[NI_MAXHOST], lport[NI_MAXSERV];
  int i;
  
  /* Check that it really is a socket and get the local address/port. */
  lAl = sizeof(lAddr);
  if(getsockname(sock, (struct sockaddr*)&lAddr, &lAl))
    return NULL;
  rAl = sizeof(lAddr);
  if(getpeername(sock, (struct sockaddr*)&rAddr, &rAl))
    return NULL;
  
  if (getnameinfo ((struct sockaddr*)&rAddr, rAl, rhost, sizeof (rhost), NULL,
	    0, NI_NUMERICHOST))
    strcpy (rhost, "unknown");
  
  if(!connections)
  {
    connections = proot();
    if(!connections)
      return NULL;
  }
  
  res = palloc(sizeof(connection_t), connections);
  if(!res)
    return NULL;
  res->sock = sock;
  res->langFd = -1;
  
  if(serv)
    res->server = pattach(serv, res);
  else
  {
    /* Find the server running on this address/port. */
    res->server = pattach(find_server((struct sockaddr*)&lAddr, lAl), res);
    if(!res->server)
    {
      if (getnameinfo ((struct sockaddr*)&lAddr, lAl, lhost, sizeof (lhost), lport,
		sizeof (lport), NI_NUMERICHOST | NI_NUMERICSERV))
      {
	strcpy (lhost, "unknown");
	strcpy (lport, "unknown");
      }
      syslog (LOG_NOTICE, "Connection from %s, but no server on %s port %s", rhost, lhost, lport);
      reply(res, "421 No service available at this address.");
      close (res->sock);
      pfree(res, connections);
      if(!errno)
	errno = EINVAL;
      return NULL;
    }
  }
  syslog (LOG_INFO, "Connection to %s from %s (id: %d)", res->server->name, rhost, res->sock);
  
  switch(lAddr.ss_family)
  {
  case AF_INET:
    res->activePort = ntohs (((struct sockaddr_in*)&lAddr)->sin_port) - 1;
    ((struct sockaddr_in*)&lAddr)->sin_port = htons (res->activePort);
    break;
  case AF_INET6:
    res->activePort = ntohs (((struct sockaddr_in6*)&lAddr)->sin6_port) - 1;
    ((struct sockaddr_in6*)&lAddr)->sin6_port = htons (res->activePort);
    break;
  }
  if(!res->server->allowLowPorts)
  {
    switch(rAddr.ss_family)
    {
    case AF_INET:
      if(ntohs (((struct sockaddr_in*)&rAddr)->sin_port) < 1024)
      {
	syslog (LOG_NOTICE, "%d: Client port < 1024", res->sock);
	reply(res, "421 Server does not accept client ports < 1024.");
	close (res->sock);
	pfree(res, connections);
	errno = EINVAL;
	return NULL;
      }
      break;
    case AF_INET6:
      if(ntohs (((struct sockaddr_in6*)&rAddr)->sin6_port) < 1024)
      {
	syslog (LOG_NOTICE, "%d: Client port < 1024", res->sock);
	reply(res, "421 Server does not accept client ports < 1024.");
	close (res->sock);
	pfree(res, connections);
	errno = EINVAL;
	return NULL;
      }
    }
  }
  res->lDataAddr = lAddr;
  res->rDataAddr = rAddr;
  res->dataSock = -1;
  res->fileFd = -1;
  res->mlstTags = tfType | tfUnique | tfModify | tfPerm | tfSize;
  
  i = -1;
  setsockopt (sock, SOL_SOCKET, SO_OOBINLINE, &i, sizeof (i));
  i = IPTOS_LOWDELAY;
  setsockopt (sock, IPPROTO_IP, IP_TOS, &i, sizeof (i));
  
  /* Add it to monitored fds and welcome the user. */
  add_read_fd(sock, control_handler, res);
  reply(res, "220 Service available at %s (%s).", res->server->name, PACKAGE_STRING);
  
  return res;
}

void disconnected(connection_t *conn)
{
  conn->spontQuit = 0;
  close_data_connection (conn);
  remove_read_fd(conn->sock);
  syslog (LOG_INFO, "%d: Disconnected", conn->sock);
#ifdef HAVE_LIBSSL
  if (conn->ssl)
  {
    SSL_shutdown (conn->ssl);
    SSL_free (conn->ssl);
  }
  if (conn->sslData)
  {
    SSL_shutdown (conn->sslData);
    SSL_free (conn->sslData);
  }
#endif
  close(conn->sock);
  if (conn->langFd >= 0)
    close_shared (conn->langFd);
  pfree(conn, connections);
}

void quit_all_connections(const char *str)
{
  connection_t *conn;
  
  if (!connections)
    return;
  
  for(conn = pchild(connections, NULL); conn; conn = pchild(connections, conn))
    reply_spont(conn, str, 1);
}

void disconnect_all_connections(void)
{
  connection_t *conn;
  
  if (!connections)
    return;
  
  while((conn = pchild(connections, NULL)))
    disconnected(conn);
}

void check_idle(void)
{
  connection_t *conn, *nconn;
  time_t currtime;
  int maxIdle;
  
  if(!connections)
    return;
  
  time(&currtime);
  
  for(conn = pchild(connections, NULL); conn; conn = nconn)
  {
    nconn = pchild(connections, conn);
    
    if (conn->sleepUntil && currtime >= conn->sleepUntil)
    {
      conn->sleepUntil = 0;
      run_commands (conn);
    }
    
    if(conn->user)
      maxIdle = conn->user->maxIdle;
    else
      maxIdle = conn->server->maxIdle;
    
    if(maxIdle && !conn->working)
    {
      if(currtime > conn->lastCommand + maxIdle * 2)
	disconnected(conn);
      else if(currtime > conn->lastCommand + maxIdle && !conn->spontQuit)
	reply_spont(conn, "421 Connection timed out.", 1);
    }
  }
}

void reply(connection_t *conn, const char *format, ...)
{
  va_list ap;
  char *str;
  int doFree = 0, isLastLine = 1;
  
  if (conn->langFd >= 0 && format)
    format = scan_strings (conn->langFd, format);
  
  if (format)
  {
    va_start (ap, format);
#ifdef HAVE_VASPRINTF
    vasprintf (&str, format, ap);
    doFree = 1;
#else
    str = talloc (4097);
    if (str)
      vsnprintf (str, 4096, format, ap);
#endif
    va_end (ap);
    
    if (!str)
    {
      str = tstring (format);
      doFree = 0;
    }
    
    if (str)
      isLastLine = (str[0] != ' ' && str[3] == ' ');
    
    // Change the numeric code if needed.
    if (str && conn->spontReply && str[0] != ' ')
    {
      str[3] = '-';
      if (conn->spontQuit)
	strncpy (str, conn->spontReply, 3);
      else
	strncpy (conn->spontReply, str, 3);
    }
    send_telnet(conn, str, 1);
  }
  
  // Only send spontaneous reply after last line of a previous message.
  if (conn->spontReply && (!format || isLastLine))
  {
    send_telnet(conn, conn->spontReply, 1);
    pfree(conn->spontReply, conn);
    conn->spontReply = NULL;
  }
  
  time(&conn->lastCommand);
  
#if HAVE_VASPRINTF
  if(doFree)
    free(str);
#endif
}

void reply_spont(connection_t *conn, const char *str, int closeConn)
{
  pfree (conn->spontReply, conn); // TODO: Append?
  conn->spontReply = pstring(str, conn);
  if(closeConn)
    conn->spontQuit = 1;
}

#define CHECK_VEC_SSP \
      if (cvec >= svec - 1) \
      { \
	svec *= 2; \
	vecs = trealloc (vecs, sizeof (struct iovec) * svec); \
	if (!vecs) \
	  return; \
      } \
      if (ssp < sp) \
      { \
	vecs[cvec].iov_base = (void*)ssp; \
	vecs[cvec++].iov_len = sp - ssp; \
      }

void send_telnet(connection_t *conn, const char *str, int newline)
{
  const char *sp, *ssp;
  struct msghdr msg = {};
  struct iovec *vecs = talloc (sizeof (struct iovec) * 4);
  int cvec = 0, svec = 4;
#ifdef HAVE_LIBSSL
  int i;
#endif
  
  if (!vecs)
    return; // TODO
  
  syslog (LOG_DEBUG, "%d: SEND: %s", conn->sock, str);
  
  for (ssp = sp = str; *sp; sp++)
  {
    switch (*(unsigned char*)sp)
    {
    case '\r':
      if (*++sp != '\n')
      {
	CHECK_VEC_SSP;
	vecs[cvec].iov_base = (void*)"";
	vecs[cvec++].iov_len = 1;
	ssp = sp--;
      }
      break;
    case 255:
      CHECK_VEC_SSP;
      vecs[cvec].iov_base = (void*)"\xFF";
      vecs[cvec++].iov_len = 1;
      ssp = sp;
      break;
    default:
      break;
    }
  }
  CHECK_VEC_SSP;
  if (newline)
  {
    vecs[cvec].iov_base = (void*)"\r\n";
    vecs[cvec++].iov_len = 2;
  }
#ifdef HAVE_LIBSSL
  if (conn->ssl)
  {
    // Presumably smart enough to handle if they get lines
    // in multiple packets.
    for (i = 0; i < cvec; i++)
      SSL_write (conn->ssl, vecs[i].iov_base, vecs[i].iov_len);
  }
  else
#endif
  {
    msg.msg_iov = vecs;
    msg.msg_iovlen = cvec;
    sendmsg (conn->sock, &msg, 0);
  }
}

void control_handler(int sock, void *user, int urgent)
{
  connection_t *conn = user;
  int l;
  char *buffp;
  
  if(urgent)
  {
    char *tmpBuf = talloc(1024);
    
    /*
     * Discard all urgent data after telnet parsing.
     * TODO Might this block if no OOB data available? Not that we should've
     * TODO been called in that case.
     */
    if(tmpBuf)
    {
      l = recv(sock, tmpBuf, 1024, MSG_OOB);
      if(l > 0)
	parse_telnet(conn, tmpBuf, l);
    }
    return;
  }
  
  if(conn->readBufferLen >= sizeof(conn->readBuffer) - 1)
  {
    /* Someone sent us a line longer than we can handle. Shame on them! */
    syslog (LOG_NOTICE, "%d: Input line overflow.", conn->sock);
    conn->overFlow = 1;
    conn->readBufferLen = 0;
  }
  
  /* Append what we read. */
  buffp = conn->readBuffer + conn->readBufferLen;
#ifdef HAVE_LIBSSL
  if (conn->ssl)
  {
    l = 0;
    switch (conn->sslState)
    {
    case 0:
      l = SSL_accept (conn->ssl);
      if (l == 1)
      {
	syslog (LOG_INFO, "%d: Successfully negotiated TLS.", conn->sock);
	conn->sslState = 1;
	if (conn->server->verifyTLS && conn->server->loginTLS)
	{
	  X509 *clientCert = SSL_get_peer_certificate (conn->ssl);
	  X509_NAME *clientName = X509_get_subject_name (clientCert);
	  char buf[100];
	  
	  if (X509_NAME_get_text_by_NID (clientName, NID_commonName, buf, 100) > 0)
	  {
	    conn->user = find_user (conn->server, buf, 1, conn);
	    if(conn->user)
	    {
	      if (conn->user->allowSecLogin)
	      {
		authenticated (conn);
		user_setup_environ (conn->user, 1);
		return;
	      }
	      pfree (conn->user, conn);
	      conn->user = NULL;
	    }
	    syslog (LOG_NOTICE, "%d: No user for CN \"%s\".", conn->sock, buf);
	  }
	  else
	    syslog (LOG_NOTICE, "%d: No CN in certificate.", conn->sock);
	  if (conn->server->tlsNoLogout)
	  {
	    reply (conn, "421 Failed to find a user for your certificate.");
	    disconnected (conn);
	  }
	}
      }
      else if (l)
      {
	if (SSL_get_error (conn->ssl, l) != SSL_ERROR_WANT_READ)
	{
	  syslog (LOG_NOTICE, "%d: Error in TLS negotiation: %s.", conn->sock,
		ERR_reason_error_string (ERR_get_error ()));
	  disconnected (conn);
	}
      }
      else
      {
	reply (conn, "421 TLS negotiation failed.");
	disconnected (conn);
      }
      return;
    case 1:
      l = SSL_read (conn->ssl, buffp, sizeof (conn->readBuffer) -
	    conn->readBufferLen);
      break;
    case 2:
      l = SSL_shutdown (conn->ssl);
      if (l == 1)
      {
	SSL_free (conn->ssl);
	conn->ssl = NULL;
      }
      else if (l && SSL_get_error (conn->ssl, l) != SSL_ERROR_WANT_READ)
      {
	syslog (LOG_NOTICE, "%d: Error in TLS shutdown: %s.", conn->sock,
	      ERR_reason_error_string (ERR_get_error ()));
	disconnected (conn);
      }
      return;
    }
  }
  else
#endif
    l = read (sock, buffp, sizeof (conn->readBuffer) - conn->readBufferLen);
  if(l < 0)
  {
    /* Error */
#ifdef HAVE_LIBSSL
    if (conn->ssl)
    {
      if (SSL_get_error (conn->ssl, l) != SSL_ERROR_WANT_READ)
      {
	syslog (LOG_NOTICE, "%d: TLS read error: %s.", conn->sock,
	      ERR_reason_error_string (ERR_get_error ()));
	disconnected (conn);
      }
    }
    else
#endif
      if(errno != EINTR)
      {
	syslog (LOG_NOTICE, "%d: control_handler: read: %m", conn->sock);
	disconnected (conn);
      }
    return;
  }
  if(!l)
  {
    syslog (LOG_NOTICE, "%d: Client suddenly disconnected.", conn->sock);
    /* EOF == they disconnected. */
    disconnected(conn);
    return;
  }
  
  if(buffp > conn->readBuffer && *(buffp - 1) == '\r')
  {
    buffp--;
    l++;
  }
  
  /* Parse telnet stuff. */
  conn->readBufferLen += parse_telnet(conn, buffp, l);
  
  if (!conn->sleepUntil) // Will be run by check_idle otherwise.
    run_commands (conn);
}

void run_commands (connection_t *conn)
{
  char *buffp;
  
  /* Check for line breaks. */
  while((buffp = memchr(conn->readBuffer, 0, conn->readBufferLen)))
  {
    if(conn->overFlow)
    {
      /* We can't handle this. Beginning of line is already discarded. */
      conn->overFlow = 0;
      reply(conn, "500 Line too long.");
    }
    else
    {
      syslog (LOG_DEBUG, "%d: RECV: %s", conn->sock, conn->readBuffer);
      if(handle_command(conn, conn->readBuffer))
	return;
    }
    
    /* Discard line */
    conn->readBufferLen -= ++buffp - conn->readBuffer;
    memmove(conn->readBuffer, buffp, conn->readBufferLen);
  }
}

int parse_telnet(connection_t *conn, char *buff, int len)
{
  unsigned char *curr, *end = (unsigned char*)buff + len;
  unsigned char retBuf[3];
  
  /*
   * Telnet command handling. Answer all DO with WON'T, and all WILL with DON'T
   * for now. Remove all null characters and put nulls at linebreaks.
   */
  for(curr = (unsigned char*)buff; curr < end;)
  {
    switch(conn->telnetMode)
    {
    case NUL:
      switch(*curr)
      {
      case NUL:
	if(conn->telnetFlags & tfCR)
	{
	  *curr = '\r';
	  if(conn->telnetFlags & tfSubOpt)
	  {
#if 0
	    if(conn->suboptLen < 100)
	      conn->subopt[conn->suboptLen++] = *curr;
#endif
	    memmove(curr, curr + 1, --end - curr);
	  }
	  else
	    curr++;
	}
	else
	  memmove(curr, curr + 1, --end - curr);
	conn->telnetFlags &= ~tfCR;
	break;
      case '\n':
	if(conn->telnetFlags & tfCR)
	  *curr = 0;
	conn->telnetFlags &= ~tfCR;
	if(conn->telnetFlags & tfSubOpt)
	{
#if 0
	  if(conn->suboptLen < 100)
	    conn->subopt[conn->suboptLen++] = *curr;
#endif
	  memmove(curr, curr + 1, --end - curr);
	}
	else
	  curr++;
	break;
      case '\r':
	conn->telnetFlags |= tfCR;
	memmove(curr, curr + 1, --end - curr);
	/*
	 * CR will be ignored unless followed by either NL or NUL, but that
	 * what the telnet rfc says a CR must be followed by so.
	 */
	break;
      case IAC:
	conn->telnetFlags &= ~tfCR;
	conn->telnetMode = IAC;
	memmove(curr, curr + 1, --end - curr);
	break;
      default:
	conn->telnetFlags &= ~tfCR;
	curr++;
	break;
      }
      break;
    case IAC:
      switch(*curr)
      {
      case IAC:
	// Quoted character 255
	if(conn->telnetFlags & tfSubOpt)
	{
#if 0
	  if(conn->suboptLen < 100)
	    conn->subopt[conn->suboptLen++] = *curr;
#endif
	  memmove(curr, curr + 1, --end - curr);
	}
	else
	  curr++;
	conn->telnetMode = NUL;
	break;
      case SB:
	memmove(curr, curr + 1, --end - curr);
	conn->telnetMode = SB;
	break;
      case SE:
	conn->telnetMode = SE;
	memmove(curr, curr + 1, --end - curr);
	break;
      case WILL:
      case WONT:
      case DO:
      case DONT:
	conn->telnetMode = *curr - 251 + 4;
	memmove(curr, curr + 1, --end - curr);
	break;
      default:
	if(*curr >= 240)
	  memmove(curr, curr + 1, --end - curr);
	else
	  curr++;
	conn->telnetMode = NUL;
	break;
      }
      break;
    case SB:
      syslog (LOG_NOTICE, "%d: Unnegotiated suboption.", conn->sock);
      conn->telnetMode = NUL;
      if (conn->telnetFlags & tfSubOpt)
	syslog (LOG_NOTICE, "%d: Nested suboption.", conn->sock);
      conn->telnetFlags |= tfSubOpt;
      conn->sb = *curr;
      memmove (curr, curr + 1, --end - curr);
      break;
    case SE:
      conn->telnetMode = NUL;
      if (!(conn->telnetFlags & tfSubOpt))
	syslog (LOG_NOTICE, "%d: Suboption end without suboption start.",
	      conn->sock);
      else if(*curr == conn->sb)
      {
	conn->telnetFlags &= ~tfSubOpt;
#if 0
	conn->suboptLen = 0;
#endif
      }
      else
	syslog (LOG_NOTICE, "%d: Wrong suboption ended, got %d expected %d.",
	      conn->sock, *curr, conn->sb);
      memmove (curr, curr + 1, --end - curr);
      break;
    case WILL:
    case DO:
      retBuf[0] = IAC;
      retBuf[1] = (conn->telnetMode == WILL? DONT : WONT);
      retBuf[2] = *curr;
      write(conn->sock, retBuf, 3);
      conn->telnetMode = NUL;
      memmove (curr, curr + 1, --end - curr);
      break;
    case WONT:
    case DONT:
      conn->telnetMode = NUL;
      memmove (curr, curr + 1, --end - curr);
      break;
    }
  }
  return (char*)end - buff;
}

void authenticated(connection_t *conn)
{
  conn->authed = 1;
  if(conn->user)
  {
    if (conn->email)
      syslog (LOG_INFO, "%d: User %s (email \"%s\") logged in.", conn->sock,
	    conn->user->name, conn->email);
    else
      syslog (LOG_INFO, "%d: User %s logged in.", conn->sock, conn->user->name);
    if(conn->user->chroot)
      set_root(conn->user->chroot);
    if(conn->user->home)
      set_cwd("~", conn->user->home);
  }
  else
    syslog (LOG_NOTICE, "%d: Unknown user logged in.", conn->sock);
}

void unauthenticated(connection_t *conn)
{
  conn->authed = 0;
  pfree(conn->user, conn);
  conn->user = NULL;
  pfree(conn->email, conn);
  conn->email = NULL;
}

int close_data_connection (connection_t *conn)
{
  int res = 0;
  
  if (conn->filePath)
  {
    time_t now;
    char rhost[NI_MAXHOST];
    int al;
    
    switch (conn->rDataAddr.ss_family)
    {
    case AF_INET:
      al = sizeof (struct sockaddr_in);
      break;
    case AF_INET6:
      al = sizeof (struct sockaddr_in6);
      break;
    default:
      al = sizeof (conn->rDataAddr);
      break;
    }
    
    if (getnameinfo ((struct sockaddr*)&conn->rDataAddr, al, rhost,
	      sizeof (rhost), NULL, 0, NI_NUMERICHOST))
      strcpy (rhost, "unknown");
    
    time (&now);
    syslog (LOG_INFO, "%d: %.24s %.0f %s %llu %s %c - %c %c %s FTP 0 * %c",
	  conn->sock, ctime (&now), difftime (now, conn->transferStart), rhost,
	  conn->dataOffset, conn->filePath,
	  conn->type == ttText? 'a' : 'b', conn->sending? 'o' : 'i',
	  conn->user && conn->user->anonymous? 'a' : 'r', // TODO might be server user.
	  conn->email? conn->email :
	  conn->user? conn->user->name : "unknown",
	  conn->dataOffset == conn->dataLen? 'c' : 'i');
    pfree (conn->filePath, conn);
    conn->filePath = NULL;
  }
  conn->passive = 0;
  conn->passiveAccepted = 0;
  conn->working = 0;
  if(conn->data)
  {
    if(conn->fileFd >= 0)
      munmap(conn->data, conn->dataLen);
    else
      pfree(conn->data, conn);
    conn->data = NULL;
  }
  conn->dataOffset = 0;
  conn->dataLen = 0;
  if(conn->fileFd >= 0)
  {
    close(conn->fileFd);
    conn->fileFd = -1;
  }
#ifdef HAVE_LIBSSL
  if (conn->sslData)
  {
    SSL_shutdown (conn->sslData);
    SSL_free (conn->sslData);
    conn->sslData = NULL;
  }
#endif
  if(conn->dataSock >= 0)
  {
    close(conn->dataSock);
    remove_read_fd(conn->dataSock);
    remove_write_fd(conn->dataSock);
    conn->dataSock = -1;
    res = 1;
  }
  if(conn->spontQuit)
  {
    conn->spontQuit = 0;
    disconnected(conn);
    return -1;
  }
  return res;
}

void passive_acceptor(int sock, void *user, int urgent)
{
  connection_t *conn = user;
  struct sockaddr_storage addr;
  int al = sizeof(addr);
  int i;
  
  set_locale (conn->currLang);
  remove_read_fd(sock);
  conn->dataSock = accept(sock, (struct sockaddr*)&addr, &al);
  if(conn->dataSock == -1)
  {
    reply(conn, "451 Acception of connection failed: %s.", strerror(errno));
    close_data_connection (conn);
    close(sock);
    return;
  }
  close(sock);
  if(!conn->server->allowForeign && !same_addr((struct sockaddr*)&addr,
	(struct sockaddr*)&conn->rDataAddr, 0))
  {
    reply(conn, "501 Foreign address not allowed.");
    close_data_connection (conn);
    return;
  }
  if(fcntl(conn->dataSock, F_SETFL, O_NONBLOCK))
  {
    reply(conn, "451 Failed to make socket nonblocking: %s.", strerror(errno));
    close_data_connection (conn);
    return;
  }
  i = IPTOS_THROUGHPUT;
  setsockopt(conn->dataSock, IPPROTO_IP, IP_TOS, &i, sizeof(i));
  conn->passiveAccepted = 1;
  if(conn->working && open_data_connection(conn))
  {
    reply(conn, "421 %s.", strerror(errno));
    close_data_connection (conn);
  }
}

int open_data_connection(connection_t *conn)
{
  int i;
  
  if(!xferBuffer)
  {
    xferBuffer = palloc(xferBufferSize, NULL);
    if(!xferBuffer)
      return -1;
  }
  
  if(conn->passive)
  {
    if(!conn->passiveAccepted)
      return 0;
  }
  else
  {
    if(conn->dataSock >= 0)
    {
      close(conn->dataSock);
      remove_read_fd(conn->dataSock);
      remove_write_fd(conn->dataSock);
    }
    conn->dataSock = socket(conn->rDataAddr.ss_family, SOCK_STREAM,
	  IPPROTO_TCP);
    if(conn->dataSock < 0)
      return -1;
    
    if(fcntl(conn->dataSock, F_SETFL, O_NONBLOCK))
    {
      close(conn->dataSock);
      conn->dataSock = -1;
      return -1;
    }
    
    super_privs(0);
    
    switch(conn->rDataAddr.ss_family)
    {
    case AF_INET:
      if(conn->lDataAddr.ss_family == AF_INET)
      {
	((struct sockaddr_in*)&conn->lDataAddr)->sin_port = htons (conn->activePort);
	bind(conn->dataSock, (struct sockaddr*)&conn->lDataAddr,
	      sizeof(struct sockaddr_in));
      }
      else if(conn->activePort)
      {
	struct sockaddr_in addr = {};
	
	addr.sin_family = AF_INET;
#ifdef HAVE_STRUCT_SOCKADDR_IN_SIN_LEN
	addr.sin_len = sizeof(addr);
#endif
	addr.sin_port = htons (conn->activePort);
	bind(conn->dataSock, (struct sockaddr*)&addr, sizeof(addr));
      }
      if(connect(conn->dataSock, (struct sockaddr*)&conn->rDataAddr,
	    sizeof(struct sockaddr_in)) && errno != EINPROGRESS)
	return -1;
      break;
    case AF_INET6:
      if(conn->lDataAddr.ss_family == AF_INET6)
      {
	((struct sockaddr_in6*)&conn->lDataAddr)->sin6_port = htons (conn->activePort);
	bind(conn->dataSock, (struct sockaddr*)&conn->lDataAddr, 
	      sizeof(struct sockaddr_in6));
      }
      else if(conn->activePort)
      {
	struct sockaddr_in6 addr = {};
	
	addr.sin6_family = AF_INET6;
#ifdef HAVE_STRUCT_SOCKADDR_IN_SIN_LEN
	addr.sin6_len = sizeof(addr);
#endif
	addr.sin6_port = htons (conn->activePort);
	bind(conn->dataSock, (struct sockaddr*)&addr, sizeof(addr));
      }
      if(connect(conn->dataSock, (struct sockaddr*)&conn->rDataAddr,
	    sizeof(struct sockaddr_in6)) && errno != EINPROGRESS)
	return -1;
      break;
    default:
      errno = EINVAL;
      return -1;
    }
  }
  i = IPTOS_THROUGHPUT;
  setsockopt(conn->dataSock, IPPROTO_IP, IP_TOS, &i, sizeof(i));
  i = xferBufferSize;
  setsockopt(conn->dataSock, SOL_SOCKET, SO_SNDLOWAT, &i, sizeof(i));
  conn->transferStart = 0;
#ifdef HAVE_LIBSSL
  if (conn->ssl)
  {
    conn->sslDataState = 0;
    if (conn->protLevel)
    {
      SSL *ssl;
      BIO *bio;
      
      ssl = SSL_new (sslCtx);
      if (!ssl)
      {
	reply (conn, "431 TLS initialisation failed.");
	close_data_connection (conn);
	return 0;
      }
      SSL_set_options (ssl, conn->server->tlsOptions);
      if (!SSL_use_certificate (ssl, conn->server->tlsCert))
      {
	syslog (LOG_ERR, "Failed to load certificate file: %s",
	      ERR_reason_error_string (ERR_get_error ()));
	reply (conn, "431 TLS initialisation failed.");
	SSL_free (ssl);
	close_data_connection (conn);
	return 0;
      }
      if (!SSL_use_PrivateKey (ssl, conn->server->tlsKey))
      {
	syslog (LOG_ERR, "Failed to load private key: %s",
	      ERR_reason_error_string (ERR_get_error ()));
	reply (conn, "431 TLS initialisation failed.");
	SSL_free (ssl);
	close_data_connection (conn);
	return 0;
      }
      if (conn->server->verifyTLS)
      {
	STACK_OF(X509_NAME) *caStack;
	
	SSL_set_verify (ssl, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
	      NULL);
	
	caStack = sk_X509_new_null ();
	SSL_add_dir_cert_subjects_to_stack (caStack, sslCertsPath);
	SSL_set_client_CA_list (ssl, caStack);
      }
      bio = BIO_new_socket (conn->dataSock, BIO_NOCLOSE);
      if (!bio)
      {
	SSL_free (ssl);
	reply (conn, "431 TLS initialisation failed.");
	close_data_connection (conn);
	return 0;
      }
      BIO_set_nbio (bio, 1);
      SSL_set_bio (ssl, bio, bio);
      conn->sslData = ssl;
      add_read_fd (conn->dataSock, data_receiver, conn);
      return 0;
    }
  }
#endif
  if (conn->sending)
  {
    /*
     * Block on sending if forking connections. We should get EINTR on SIGURG,
     * and since we set SNDLOWAT to the xfer buffer size it should only apply
     * to mmap() sends.
     */
    if (forkConnections)
      fcntl (conn->dataSock, F_SETFL, 0);
    add_write_fd (conn->dataSock, data_sender, conn);
  }
  else
    add_read_fd (conn->dataSock, data_receiver, conn);
  return 0;
}

int check_secure_data (connection_t *conn)
{
#ifdef HAVE_LIBSSL
  int l;
  
  if (conn->sslData)
  {
    switch (conn->sslDataState)
    {
    case 0:
      l = SSL_accept (conn->sslData);
      if (l == 1)
      {
	syslog (LOG_INFO, "%d: Successfully negotiated data TLS.", conn->sock);
	conn->sslDataState = 1;
	if (conn->sending)
	{
	  remove_read_fd (conn->dataSock);
	  add_write_fd (conn->dataSock, data_sender, conn);
	}
	if (conn->server->verifyTLS)
	{
	  X509 *clientCert = SSL_get_peer_certificate (conn->ssl);
	  X509 *clientDataCert = SSL_get_peer_certificate (conn->sslData);
	  
	  if (X509_cmp (clientCert, clientDataCert))
	  {
	    syslog (LOG_NOTICE, "%d: Control/data certificate mismatch.", conn->sock);
	    reply (conn, "535 Must use same certificate on control and data connections.");
	    close_data_connection (conn);
	  }
	}
      }
      else if (l)
      {
	if (SSL_get_error (conn->sslData, l) != SSL_ERROR_WANT_READ)
	{
	  syslog (LOG_NOTICE, "%d: Error in data TLS negotiation: %s.", conn->sock,
		ERR_reason_error_string (ERR_get_error ()));
	  close_data_connection (conn);
	}
      }
      else
      {
	reply (conn, "431 Data TLS negotiation failed.");
	close_data_connection (conn);
      }
      return 1;
    case 2:
      l = SSL_shutdown (conn->sslData);
      if (l == 1)
	close_data_connection (conn);
      else if (l && SSL_get_error (conn->sslData, l) != SSL_ERROR_WANT_READ)
      {
	syslog (LOG_NOTICE, "%d: Error in TLS shutdown: %s.", conn->sock,
	      ERR_reason_error_string (ERR_get_error ()));
	close_data_connection (conn);
      }
      return 1;
    }
  }
#endif
  return 0;
}

void data_receiver(int sock, void *user, int urgent)
{
  connection_t *conn = user;
  int rl, wl;
  char *bp1, *bp2;
  
  set_locale (conn->currLang);
  if (conn->protLevel && check_secure_data (conn))
    return;
  
  if (!conn->transferStart)
    time (&conn->transferStart);
  
#ifdef HAVE_LIBSSL
  if (conn->sslData)
    rl = SSL_read (conn->sslData, xferBuffer, xferBufferSize);
  else
#endif
    rl = recv(sock, xferBuffer, xferBufferSize, 0);
  if(rl < 0)
  {
    reply(conn, "451 socket error: %s.",
#ifdef HAVE_LIBSSL
	  conn->sslData? ERR_reason_error_string (ERR_get_error ()) :
#endif
	  strerror(errno));
    close_data_connection (conn);
    return;
  }
  if(!rl)
  {
    /* EOF */
    if (!conn->dataLen)
      conn->dataLen = conn->dataOffset;
    reply(conn, "226 Transfer complete.");
    close_data_connection (conn);
    return;
  }
  if(conn->type == ttText)
  {
    wl = 0;
    for(bp1 = xferBuffer; (bp2 = memchr(bp1, '\r', rl)); bp1 = bp2 + 1)
    {
      wl = write(conn->fileFd, bp1, bp2 - bp1);
      if(wl == -1)
	break;
      conn->dataOffset += wl + 1;
      rl -= wl + 1;
    }
    if(rl && wl != -1)
    {
      wl = write(conn->fileFd, bp1, rl);
      conn->dataOffset += rl;
    }
  }
  else
  {
    wl = write(conn->fileFd, xferBuffer, rl);
    conn->dataOffset += rl;
  }
  if(wl == -1)
  {
    if(errno == ENOSPC)
      reply(conn, "452 Out of disk space.");
    else if(errno == EDQUOT)
      reply(conn, "552 Quota exceeded.");
    else
      reply(conn, "451 write failed: %s.", strerror(errno));
    close_data_connection (conn);
  }
}

void data_sender(int sock, void *user, int urgent)
{
  connection_t *conn = user;
  int rl, wl;
  char *bp1, *bp2;
  
  set_locale (conn->currLang);
  if (!conn->transferStart)
    time (&conn->transferStart);
  
  rl = sizeof(errno);
  if(!getsockopt(sock, SOL_SOCKET, SO_ERROR, &errno, &rl))
  {
    if(errno)
    {
      reply(conn, "451 socket error: %s.", strerror(errno));
      close_data_connection (conn);
      return;
    }
  }
  
#ifdef HAVE_LIBSSL
  if (conn->ssl)
    urgData = 1; // Force exit after one loop.
#endif
  
  if(conn->data)
  {
    do
    {
      if(conn->dataOffset == conn->dataLen)
      {
	/* EOF */
	reply(conn, "226 Transfer complete.");
	close_data_connection (conn);
	return;
      }
#ifdef HAVE_LIBSSL
      if (conn->sslData)
	wl = SSL_write (conn->sslData, (char*)conn->data + conn->dataOffset,
	      conn->dataLen - conn->dataOffset);
      else
#endif
	wl = write(conn->dataSock, (char*)conn->data + conn->dataOffset,
	      conn->dataLen - conn->dataOffset);
      if(wl >= 0)
	conn->dataOffset += wl;
      else
      {
#ifdef HAVE_LIBSSL
	if (conn->sslData)
	{
	  wl = SSL_get_error (conn->sslData, wl);
	  if (wl != SSL_ERROR_WANT_READ && wl != SSL_ERROR_WANT_WRITE)
	  {
	    reply (conn, "451 Write to socket failed: %s.", 
		  ERR_reason_error_string (ERR_get_error ()));
	    close_data_connection (conn);
	  }
	}
	else
#endif
	  if (errno != EINTR && errno != EAGAIN)
	  {
	    reply (conn, "451 Write to socket failed: %s.", strerror(errno));
	    close_data_connection (conn);
	  }
	return;
      }
    } while(forkConnections && !urgData);
  }
  else if(conn->type == ttText)
  {
    do
    {
      rl = read(conn->fileFd, xferBuffer, xferBufferSize);
      if(rl == -1)
      {
	reply(conn, "451 read failed: %s.", strerror(errno));
	close_data_connection (conn);
	return;
      }
      if(!rl)
      {
	/* EOF */
	conn->dataLen = conn->dataOffset;
	reply(conn, "226 Transfer complete.");
	close_data_connection (conn);
	return;
      }
      conn->dataOffset += rl;
      wl = 0;
      for(bp1 = xferBuffer; (bp2 = memchr(bp1, '\n', rl)); bp1 = bp2 + 1)
      {
#ifdef HAVE_LIBSSL
	if (conn->sslData)
	  wl = SSL_write (conn->sslData, bp1, bp2 - bp1);
	else
#endif
	  wl = write (sock, bp1, bp2 - bp1);
	if(wl < 0)
	  break;
	rl -= wl + 1;
#ifdef HAVE_LIBSSL
	if (conn->sslData)
	  wl = SSL_write (conn->sslData, "\r\n", 2);
	else
#endif
	  wl = write (sock, "\r\n", 2);
	if(wl < 0)
	  break;
      }
      if (rl && wl >= 0)
      {
#ifdef HAVE_LIBSSL
	if (conn->sslData)
	  wl = SSL_write (conn->sslData, bp1, rl);
	else
#endif
	  wl = write(sock, bp1, rl);
      }
      if(wl < 0)
      {
#ifdef HAVE_LIBSSL
	if (conn->sslData)
	{
	  wl = SSL_get_error (conn->sslData, wl);
	  if (wl != SSL_ERROR_WANT_READ && wl != SSL_ERROR_WANT_WRITE)
	  {
	    reply (conn, "451 Write to socket failed: %s.", 
		  ERR_reason_error_string (ERR_get_error ()));
	    close_data_connection (conn);
	  }
	  else
	    lseek (conn->fileFd, SEEK_CUR, -rl);
	}
	else
#endif
	  if (errno != EINTR && errno != EAGAIN)
	  {
	    reply (conn, "451 Write to socket failed: %s.", strerror(errno));
	    close_data_connection (conn);
	  }
	  else
	    lseek (conn->fileFd, SEEK_CUR, -rl);
	return;
      }
    } while(forkConnections && !urgData);
  }
  else
  {
    do
    {
      rl = read(conn->fileFd, xferBuffer, xferBufferSize);
      if(rl == -1)
      {
	reply(conn, "451 read failed: %s.", strerror(errno));
	close_data_connection (conn);
	return;
      }
      if(!rl)
      {
	/* EOF */
	conn->dataLen = conn->dataOffset;
	reply(conn, "226 Transfer complete.");
	close_data_connection (conn);
	return;
      }
      conn->dataOffset += rl;
#ifdef HAVE_LIBSSL
      if (conn->sslData)
	wl = SSL_write (conn->sslData, xferBuffer, rl);
      else
#endif
	wl = write(sock, xferBuffer, rl);
      if (wl < 0)
      {
#ifdef HAVE_LIBSSL
	if (conn->sslData)
	{
	  wl = SSL_get_error (conn->sslData, wl);
	  if (wl != SSL_ERROR_WANT_READ && wl != SSL_ERROR_WANT_WRITE)
	  {
	    reply (conn, "451 Write to socket failed: %s.", 
		  ERR_reason_error_string (ERR_get_error ()));
	    close_data_connection (conn);
	  }
	  else
	    lseek (conn->fileFd, SEEK_CUR, -rl);
	}
	else
#endif
	  if (errno != EINTR && errno != EAGAIN)
	  {
	    reply (conn, "451 Write to socket failed: %s.", strerror (errno));
	    close_data_connection (conn);
	  }
	  else
	    lseek (conn->fileFd, SEEK_CUR, -rl);
	return;
      }
      else if(wl < rl)
	lseek(conn->fileFd, SEEK_CUR, wl - rl);
    } while(forkConnections && !urgData);
  }
}
