/*****************************************************************************\
* Copyright (c) 2002 Pelle Johansson.                                         *
* All rights reserved.                                                        *
*                                                                             *
* This file is part of the moftpd package. Use and distribution of            *
* this software is governed by the terms in the file LICENCE, which           *
* should have come with this package.                                         *
\*****************************************************************************/

/* $moftpd: main.c,v 1.21.2.8 2003/09/26 21:45:23 morth Exp $ */
/* $Id$ */

#include "main.h"

#include "connection.h"
#include "server.h"
#include "utf8fs/file.h"
#include "confparse.h"
#include "memory.h"

#include "system.h"

/* TODO should probably be a linked list. */
typedef struct fdHandler
{
  fdHandlerFun_t handler;
  void *user;
} fdHandler_t;

static fd_set readSet, writeSet;
static int doFork = 1, maxFd = 0, reloadConfig = 0;
static fdHandler_t readHandlers[FD_SETSIZE], writeHandlers[FD_SETSIZE];

int debug = 0;
int running = 1, runningAsRoot;
const char *configFile = ETCDIR "/moftpd.conf";
int forkConnections = 1, skipFd0 = 0;
unsigned long long maxMmapSize = 10 * 1024 * 1024;
int urgData = 0;
int unprivUid, unprivGid;
const char *localeDir = SHAREDIR "/" PACKAGE_NAME "/locale/";
const char *localeSuffix = ".loc";
#ifdef HAVE_LIBSSL
SSL_CTX *sslCtx;
char *sslCertsPath;
#endif

extern char *optarg;
extern int optind;
extern int fakeChroot;

void sigkill(int unused)
{
  running = 0;
}

void sigquit(int unused)
{
  running = 0;
}

void sigurg(int unused)
{
  urgData = 1;
}

void sigchld(int unused)
{
  wait(NULL);
}

void sigint(int unused)
{
  int i;
  
  quit_all_connections("421 Server shutting down.");
  
  for(i = 0; i < maxFd; i++)
  {
    // A little hacky, but best way at the moment.
    if(readHandlers[i].handler == connection_accepter)
    {
      remove_read_fd(i);
      close(i);
    }
  }
}

void sighup(int unused)
{
  sigint(unused);
  reloadConfig = 1;
}

int main(int argc, char *argv[])
{
  connection_t *conn;
  char startCwd[MAXPATHLEN];
  
  signal(SIGKILL, sigkill);
  signal(SIGQUIT, sigquit);
  signal(SIGURG, sigurg);
  signal(SIGCHLD, sigchld);
  signal(SIGINT, sigint);
  signal(SIGHUP, sighup);
  signal (SIGPIPE, SIG_IGN);
  
  FD_ZERO(&readSet);
  FD_ZERO(&writeSet);
  
  setprogname (argv[0]);
  unprivUid = geteuid();
  unprivGid = getegid();
  
  runningAsRoot = !unprivUid;
  
  mem_init ();
  
#ifdef HAVE_LIBSSL
  sslCertsPath = pstring (X509_get_default_cert_dir (), NULL);
#endif
  
  argc = parse_options(argc, &argv);
  if(argc < 0)
    return 1;
  if(argc && !getenv("MOFTP_CHILD"))
    usage();
  
  /*
   * Use NDELAY since we might chroot.
   */
  openlog (getprogname (), LOG_PID | LOG_NDELAY | (debug? LOG_PERROR : 0),
	LOG_FTP);
  if (!debug)
  {
#ifdef RLIMIT_CORE
    struct rlimit rl;
    
    if (getrlimit (RLIMIT_CORE, &rl))
      rl.rlim_max = 0;
    rl.rlim_cur = 0;
    setrlimit (RLIMIT_CORE, &rl);
#endif
    setlogmask (LOG_UPTO (LOG_INFO));
  }
  
  getcwd (startCwd, sizeof (startCwd));
  
 loadConfig:
  if (reloadConfig)
  {
    reloadConfig = 0;
    chdir (startCwd);
  }
  
  if(read_config(configFile))
    return 1;
  
  if(!fakeChroot && (!runningAsRoot || !forkConnections))
  {
    syslog (LOG_WARNING, "Warning: FakeChroot turned on because %s.",
	  runningAsRoot? "ForkConnections is false" : "not running as root");
    fakeChroot = 1;
  }
  
  if(skipFd0)
  {
    conn = NULL;
    errno = 0;
  }
  else
  {
    /*
     * Check if we have a connection on stdin.
     * Not sure this is foolproof with regards to pipes, but I don't see why
     * you'd start moftpd with a pipe on stdin. find_server() only works on
     * PF_INET and PF_INET6, so you'll get no service available on pipe
     * sockets.
     */
    conn = new_connection(0, NULL);
    skipFd0 = 1;
  }
  if(conn)
  {
    /*
     * Seems we do. Means we've been charged with handling this connection
     * only. new_connection() should've added it to readSet.
     */
    forkConnections = 1; // This is in effect a forked connection.
    quit_all_servers ();
  }
  else if(errno && errno != ENOTSOCK)
  {
    /* Error! */
    return 1;
  }
  else
  {
    /* We're in daemon mode. Setup as appropriate. */
    
    create_server_sockets();
    
#ifdef HAVE_SETENV
    setenv("MOFTP_CHILD", "1", 0); /* For children we launch. */
#else
    putenv (tstring ("MOFTP_CHILD=1"));
#endif
    
    if(doFork)
    {
      switch(fork())
      {
      case 0:
	/* Daemon. Keep running. */
	break;
      case -1:
	/* Error */
	syslog(LOG_ERR, "fork: %m");
	return 1;
      default:
	/* Parent. We're done. */
	return 0;
      }
      close (0);
      close (1);
      if (!debug)
      {
        close (2);
        setsid ();
      }
    }
  }
  
  /* Main loop. */
  while (running && maxFd > 0)
  {
    fd_set rSet, wSet, eSet;
    int fdIter;
    struct timeval tv;
    
    drop_privs();
    
    rSet = readSet;
    wSet = writeSet;
    eSet = readSet;
    
    tv.tv_sec = 10;
    tv.tv_usec = 0;
    
    tfree_all();
    
    if(select(maxFd, &rSet, &wSet, &eSet, &tv) < 0)
    {
      if(errno == EINTR)
	continue;
      syslog (LOG_ERR, "select: %m");
      return 1;
    }
    urgData = 0;
    
    for(fdIter = 0; fdIter < maxFd; fdIter++)
    {
      /*
       * Cheack readSet and writeSet as well as they might have been removed
       * by an earlier handler.
       */
      if(FD_ISSET(fdIter, &eSet) && FD_ISSET(fdIter, &readSet))
      {
	/*
	 * "Exceptional state" means urgent data available on sockets. 
	 * Hopefully that's the only thing that activates it, not completely
	 * sure.
	 */
	readHandlers[fdIter].handler(fdIter, readHandlers[fdIter].user, 1);
      }
      if(FD_ISSET(fdIter, &rSet) && FD_ISSET(fdIter, &readSet))
	readHandlers[fdIter].handler(fdIter, readHandlers[fdIter].user, 0);
      if(FD_ISSET(fdIter, &wSet) && FD_ISSET(fdIter, &writeSet))
	writeHandlers[fdIter].handler(fdIter, writeHandlers[fdIter].user, 0);
    }
    check_idle();
    
    if (reloadConfig)
    {
      if (!super_privs (1))
	break;
      reloadConfig = 0;
    }
  }
  
  quit_all_servers ();
  
  if(running && reloadConfig)
    goto loadConfig;
  
  return 0;
}

int add_read_fd(int fd, fdHandlerFun_t handler, void *user)
{
  if(fd >= FD_SETSIZE || !handler)
  {
    /*
     * We're restricted by FD_SETSIZE in select() and don't allow
     * NULL handlers.
     */
    errno = EINVAL;
    return -1;
  }
  
  /* Update maxFd if appropriate. */
  if(fd >= maxFd)
    maxFd = fd + 1;
  
  FD_SET(fd, &readSet);
  readHandlers[fd].handler = handler;
  readHandlers[fd].user = user;
  
  return 0;
}

int add_write_fd(int fd, fdHandlerFun_t handler, void *user)
{
  if(fd >= FD_SETSIZE || !handler)
  {
    /*
     * We're restricted by FD_SETSIZE in select() and don't allow
     * NULL handlers.
     */
    errno = EINVAL;
    return -1;
  }
  
  /* Update maxFd if appropriate. */
  if(fd >= maxFd)
    maxFd = fd + 1;
  
  FD_SET(fd, &writeSet);
  writeHandlers[fd].handler = handler;
  writeHandlers[fd].user = user;
  
  return 0;
}

void remove_read_fd(int fd)
{
  if(fd < 0 || fd >= FD_SETSIZE)
    return;
  
  FD_CLR(fd, &readSet);
  
  /* We might be able to lower maxFd. Check! */
  if(fd == maxFd - 1)
  {
    int fdIter, newMax = -1;
    
    for(fdIter = 0; fdIter < maxFd; fdIter++)
    {
      if(FD_ISSET(fdIter, &readSet) || FD_ISSET(fdIter, &writeSet))
	newMax = fdIter;
    }
    maxFd = newMax + 1;
  }
}

void remove_write_fd(int fd)
{
  if(fd < 0 || fd >= FD_SETSIZE)
    return;
  
  FD_CLR(fd, &writeSet);
  
  /* We might be able to lower maxFd. Check! */
  if(fd == maxFd - 1)
  {
    int fdIter, newMax = -1;
    
    for(fdIter = 0; fdIter < maxFd; fdIter++)
    {
      if(FD_ISSET(fdIter, &readSet) || FD_ISSET(fdIter, &writeSet))
	newMax = fdIter;
    }
    maxFd = newMax + 1;
  }
}

int parse_options(int argc, char ***argv)
{
  int ch;
  char dbuf[4];
  
  if(getenv("MOFTPD_CONFIG"))
  {
    configFile = pstring (getenv ("MOFTP_CONFIG"), NULL);
    if(!configFile)
    {
      perror ("pstring");
      return -1;
    }
  }
  if (getenv ("MOFTPD_DEBUG"))
    debug = atoi (getenv ("MOFTPD_DEBUG"));
  
  while((ch = getopt(argc, *argv, "c:dfD")) != -1)
  {
    switch(ch)
    {
    case 'c':
      configFile = pstring(optarg, NULL);
      if(!configFile)
      {
	perror ("pstring");
	return -1;
      }
      break;
    case 'D':
      debug++;
      break;
    case 'd':
      doFork = 1;
      skipFd0 = 1;
      break;
    case 'f':
      if(!doFork)
	forkConnections = 0;
      else
	doFork = 0;
      break;
    default:
      usage();
    }
  }
  
  if(debug < 0 || debug > 999)
    debug = 999;
#ifdef HAVE_SETENV
  setenv("MOFTPD_CONFIG", configFile, 1);
  sprintf(dbuf, "%d", debug);
  setenv("MOFTPD_DEBUG", dbuf, 1);
#else
  {
    char *str = talloc (4097);
    
    if (str)
    {
      snprintf (str, 4096, "MOFTPD_CONFIG=%s", configFile);
      putenv (str);
      snprintf (str, 4096, "MOFTPD_DEBUG=%d", debug);
      putenv (str);
    }
  }
#endif
  
  *argv += optind;
  return argc - optind;
}

void usage(void)
{
  fprintf (stderr, "Usage: %s [-dDf] [-c <configfile>]\n", getprogname ());
  exit(1);
}

int super_privs(int resetRoot)
{
  if(resetRoot && set_root("/"))
    return -1;
  
  if(runningAsRoot)
  {
    seteuid(0);
    
    if(setuid(0))
      return -1;
    
    setgid(0);
  }
  
  return 0;
}

int drop_privs(void)
{
  if(runningAsRoot)
  {
    seteuid (0);
    setegid(unprivGid);
    seteuid(unprivUid);
  }
  return 0;
}

int set_uid(uid_t uid, int permanent)
{
  if(runningAsRoot)
  {
    seteuid(0);
    if (permanent)
      return setuid (uid);
    return seteuid(uid);
  }
  return 0;
}

int set_gid(gid_t gid, int permanent)
{
  if(runningAsRoot)
  {
    int euid = geteuid(), res;
    
    seteuid(0);
    if (permanent)
      res = setgid (gid);
    else
      res = setegid (gid);
    seteuid(euid);
    return res;
  }
  return 0;
}

int set_groups(int ngroups, const gid_t *groups)
{
  if(runningAsRoot)
  {
    int euid = geteuid(), res;
    
    seteuid(0);
    res = setgroups(ngroups, groups);
    seteuid(euid);
    return res;
  }
  return 0;
}

void set_locale (const char *loc)
{
  char lbuf[3];
  
  if (loc)
  {
    strncpy (lbuf, loc, 2);
    lbuf[2] = 0;
    if (!setlocale (LC_ALL, lbuf))
      setlocale (LC_ALL, "en");
  }
  else
    setlocale (LC_ALL, "en");
}

int same_addr(const struct sockaddr *a1, const struct sockaddr *a2,
      int checkPort)
{
  if(a1->sa_family != a2->sa_family)
    return 0;
  
  switch(a1->sa_family)
  {
  case AF_INET:
    if(checkPort)
    {
      if(((struct sockaddr_in*)a1)->sin_port != ((struct sockaddr_in*)a2)->
	    sin_port)
	return 0;
    }
    return !memcmp(&((struct sockaddr_in*)a1)->sin_addr, &((struct
	  sockaddr_in*)a2)->sin_addr, 4);
  case AF_INET6:
    if(checkPort)
    {
      if(((struct sockaddr_in6*)a1)->sin6_port != ((struct sockaddr_in6*)a2)->
	    sin6_port)
	return 0;
    }
    return !memcmp(&((struct sockaddr_in6*)a1)->sin6_addr, &((struct
	  sockaddr_in6*)a2)->sin6_addr, 16);
  default:
    return 0;
  }
}
