/*****************************************************************************\
* Copyright (c) 2002 Pelle Johansson.                                         *
* All rights reserved.                                                        *
*                                                                             *
* This file is part of the moftpd package. Use and distribution of            *
* this software is governed by the terms in the file LICENCE, which           *
* should have come with this package.                                         *
\*****************************************************************************/

/* $moftpd: server.c,v 1.13.2.4 2004/03/18 15:10:03 morth Exp $ */
/* $Id$ */

#include "server.h"

#include "main.h"
#include "connection.h"
#include "memory.h"
#include "defaults.h"

#include "system.h"

static void *servers;

extern int forkConnections, debug;

server_t *new_server(const char *name, server_t *parent)
{
  server_t *res;
  
  if(!servers)
  {
    servers = proot();
    if(!servers)
      return NULL;
  }
  
  res = palloc(sizeof(server_t), servers);
  
  if(!res)
    return NULL;
  
  res->name = pstring(name, res);
  if(!res->name)
  {
    pfree(res, servers);
    return NULL;
  }
  
  if(parent)
  {
    res->numPorts = parent->numPorts;
    if(res->numPorts)
    {
      res->ports = palloc(sizeof(int) * res->numPorts, res);
      if(!res->ports)
      {
	pfree(res, servers);
	return NULL;
      }
      memcpy(res->ports, parent->ports, sizeof(int) * res->numPorts);
    }
    res->passIfInvalid = parent->passIfInvalid;
    res->allowForeign = parent->allowForeign;
    res->allowLowPorts = parent->allowLowPorts;
    res->minPasvPort = parent->minPasvPort;
    res->maxPasvPort = parent->maxPasvPort;
    res->users = pattach(parent->users, res);
    res->access = pattach(parent->access, res);
    res->chroot = pattach(parent->chroot, res);
    res->allowLogin = parent->allowLogin;
    res->passwordNeeded = parent->passwordNeeded;
    res->loginFailedMsg = pattach(parent->loginFailedMsg, res);
    res->userInvalidMsg = pattach(parent->userInvalidMsg, res);
    res->passRequestMsg = pattach(parent->passRequestMsg, res);
    res->anonPassMsg = pattach(parent->anonPassMsg, res);
    res->maxIdle = parent->maxIdle;
    res->sleepOnFail = parent->sleepOnFail;
    res->maxLoginAttempts = parent->maxLoginAttempts;
    res->allowSecLogin = parent->allowSecLogin;
    res->numAliases = parent->numAliases;
    if(res->numAliases)
    {
      res->aliases = palloc(sizeof(*res->aliases) * res->numAliases, res);
      if(!res->aliases)
      {
	pfree(res, servers);
	return NULL;
      }
      memcpy(res->aliases, parent->aliases, sizeof(*res->aliases) * res->numAliases);
    }
  }
  else
  {
    res->allowForeign = defAllowForeign;
    res->allowLowPorts = defAllowLowPorts;
    res->allowUnbound = defAllowUnbound;
    res->minPasvPort = defMinPasvPort;
    res->maxPasvPort = defMaxPasvPort;
    res->passIfInvalid = defPassIfInvalid;
    res->allowLogin = defAllowLogin;
    res->allowSecLogin = defAllowSecLogin;
    res->passwordNeeded = defPasswordNeeded;
    res->maxIdle = defMaxIdle;
    res->sleepOnFail = defSleepOnFail;
    res->maxLoginAttempts = defMaxLoginAttempts;
    res->tlsOptions = defTLSOptions;
  }
  
  return res;
}

void free_server(server_t *serv)
{
  pfree(serv, servers);
}

void quit_all_servers()
{
  pfree (servers, NULL);
  servers = NULL;
}

void add_server_port(server_t *serv, int port)
{
  if(serv->numPorts++)
    serv->ports = prealloc(serv->ports, serv->numPorts * sizeof(*serv->ports));
  else
    serv->ports = palloc(sizeof(*serv->ports), serv);
  serv->ports[serv->numPorts - 1] = port;
}

void add_server_binding(server_t *serv, const char *mask)
{
  if(serv->numBindings++)
    serv->bindings = prealloc(serv->bindings, serv->numBindings *
	  sizeof(*serv->bindings));
  else
    serv->bindings = palloc(sizeof(*serv->bindings), serv);
  serv->bindings[serv->numBindings - 1] = pstring (mask, serv->bindings);
}

void server_add_alias(server_t *serv, const char *alias, const char *user)
{
  int i;
  
  for(i = 0; i < serv->numAliases; i++)
  {
    if(!strcmp(serv->aliases[i].alias, alias))
    {
      pfree(serv->aliases[i].user, serv->aliases);
      serv->aliases[i].user = pstring(user, serv->aliases);
      return;
    }
  }
  
  if(serv->numAliases++)
    serv->aliases = prealloc(serv->aliases, serv->numAliases *
	  sizeof(*serv->aliases));
  else
    serv->aliases = palloc(sizeof(*serv->aliases), serv);
  serv->aliases[serv->numAliases - 1].alias = pstring(alias, serv->aliases);
  serv->aliases[serv->numAliases - 1].user = pstring(user, serv->aliases);
}

const char *server_expand_alias(const server_t *serv, const char *name)
{
  int i;
  
  for(i = 0; i < serv->numAliases; i++)
  {
    if(!strcmp(serv->aliases[i].alias, name))
      return serv->aliases[i].user;
  }
  
  return name;
}

server_t *find_server(const struct sockaddr *addr, int addrlen)
{
  char host[NI_MAXHOST];
  int port;
  server_t *res;
  struct addrinfo hints = {}, *aires, *aicurr;
  int i;
  
  switch(addr->sa_family)
  {
  case AF_INET:
    port = ntohs (((struct sockaddr_in*)addr)->sin_port);
    break;
  case AF_INET6:
    port = ntohs (((struct sockaddr_in6*)addr)->sin6_port);
    break;
  default:
    return NULL;
  }
  
  /* First look for numeric host to avoid dns lookup timeout. */
  if(!getnameinfo(addr, addrlen, host, sizeof(host), NULL, 0,
	NI_NUMERICHOST))
  {
    res = find_named_server(host, port);
    if(res)
      return res;
  }
  
  /* Next lookup hostname. */
  if(!getnameinfo(addr, addrlen, host, sizeof(host), NULL, 0, NI_NAMEREQD))
  {
    res = find_named_server(host, port);
    if(res)
      return res;
  }
  
  /* Next lookup configured server names. */
  hints.ai_socktype = SOCK_STREAM;
  for(res = pchild(servers, NULL); res; res = pchild(servers, res))
  {
    for(i = 0; i < res->numPorts; i++)
      if(res->ports[i] == port)
	break;
    if(i == res->numPorts)
      continue;
    
    if(strchr(res->name, '*') || strchr(res->name, '?'))
      continue;
    
    aires = NULL;
    if(!getaddrinfo(res->name, NULL, &hints, &aires) && aires)
    {
      for(aicurr = aires; aicurr; aicurr = aicurr->ai_next)
      {
	if(same_addr(addr, aicurr->ai_addr, 0))
	  break;
      }
      freeaddrinfo(aires);
      if(aicurr)
	return res;
    }
  }
  
  /* Next lookup configured servers bindings. */
  for(res = pchild(servers, NULL); res; res = pchild(servers, res))
  {
    for(i = 0; i < res->numPorts; i++)
      if(res->ports[i] == port)
	break;
    if(i == res->numPorts)
      continue;
    
    for(i = 0; i < res->numBindings; i++)
    {
      if(strchr(res->bindings[i], '*') || strchr(res->bindings[i], '?'))
	continue;
      
      aires = NULL;
      if(!getaddrinfo(res->bindings[i], NULL, &hints, &aires) && aires)
      {
	for(aicurr = aires; aicurr; aicurr = aicurr->ai_next)
	{
	  if(same_addr(addr, aicurr->ai_addr, 0))
	    break;
	}
	freeaddrinfo(aires);
	if(aicurr)
	  return res;
      }
    }
  }
  
  /* Last find first server on port that allows nonbound. */
  for(res = pchild(servers, NULL); res; res = pchild(servers, res))
  {
    if(!res->allowUnbound)
      continue;
    
    for(i = 0; i < res->numPorts; i++)
      if(res->ports[i] == port)
	break;
    if(i < res->numPorts)
      return res;
  }
  
  /* No server found. */
  return NULL;
}

server_t *find_named_server(char *host, int port)
{
  server_t *res;
  int i;
  
  /* First check versus names. */
  for(res = pchild(servers, NULL); res; res = pchild(servers, res))
  {
    for(i = 0; i < res->numPorts; i++)
      if(res->ports[i] == port)
	break;
    if(i == res->numPorts)
      continue;
    
    if(!strcasecmp(host, res->name))
      return res;
  }
  
  /* Then check versus bindings. */
  for(res = pchild(servers, NULL); res; res = pchild(servers, res))
  {
    for(i = 0; i < res->numPorts; i++)
      if(res->ports[i] == port)
	break;
    if(i == res->numPorts)
      continue;
    
    for(i = 0; i < res->numBindings; i++)
    {
      if(match_pattern(res->bindings[i], host))
	return res;
    }
  }
  
  /* No server found. */
  return NULL;
}

int create_server_sockets(void)
{
  server_t *serv;
  int *allPorts = NULL;
  int numPorts = 0, maxNumPorts = 0;
  int isAll, i, j, sock;
  struct sockaddr_in addr4 = {};
  struct sockaddr_in6 addr6 = {};
  
  for(serv = pchild(servers, NULL); serv; serv = pchild(servers, serv))
  {
    isAll = 1; /* TODO Implement notAll */
    
    if(serv->allowUnbound)
      isAll = 1;
    else for(i = 0; i < serv->numBindings; i++)
    {
      if(strchr(serv->bindings[i], '*') || strchr(serv->bindings[i], '?'))
      {
	/* Wildcards, have to listen on all. */
	isAll = 1;
	break;
      }
    }
    
    if(isAll)
    {
      for(i = 0; i < serv->numPorts; i++)
      {
	for(j = 0; j < numPorts; j++)
	{
	  if(allPorts[j] == serv->ports[i])
	    break;
	}
	if(j < numPorts)
	  continue;
	if(numPorts < maxNumPorts)
	  allPorts[numPorts++] = serv->ports[i];
	else if(maxNumPorts)
	{
	  maxNumPorts *= 2;
	  allPorts = trealloc(allPorts, maxNumPorts * sizeof(*allPorts));
	  if(!allPorts)
	  {
	    syslog (LOG_ERR, "trealloc: %m");
	    return -1;
	  }
	  allPorts[numPorts++] = serv->ports[i];
	}
	else
	{
	  maxNumPorts = 10; /* Start high, only temporary. */
	  allPorts = talloc(maxNumPorts * sizeof(*allPorts));
	  if(!allPorts)
	  {
	    syslog (LOG_ERR, "talloc: %m");
	    return -1;
	  }
	  allPorts[numPorts++] = serv->ports[i];
	}
      }
    }
  }
  
  for(i = 0; i < numPorts; i++)
  {
    sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(sock < 0)
      syslog (LOG_ERR, "socket: %m");
    else
    {
      j = -1;
      setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &j, sizeof(j));
      
      addr4.sin_family = AF_INET;
#ifdef HAVE_STRUCT_SOCKADDR_IN_SIN_LEN
      addr4.sin_len = sizeof(addr4);
#endif
      addr4.sin_port = htons (allPorts[i]);
      
      if(bind(sock, (struct sockaddr*)&addr4, sizeof(addr4)))
      {
	syslog (LOG_ERR, "bind: %m");
	close(sock);
      }
      else if(listen(sock, 128))
      {
	syslog (LOG_ERR, "listen: %m");
	close(sock);
      }
      else
	add_read_fd(sock, connection_accepter, NULL);
    }
    
    sock = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
    if(sock < 0)
      syslog (LOG_ERR, "socket: %m");
    else
    {
      j = -1;
      setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &j, sizeof(j));
#ifdef IPV6_V6ONLY
      // Only IPv6 please.
      setsockopt (sock, IPPROTO_IPV6, IPV6_V6ONLY, &j, sizeof (j));
#endif
      
      addr6.sin6_family = AF_INET6;
#ifdef HAVE_STRUCT_SOCKADDR_IN_SIN_LEN
      addr6.sin6_len = sizeof(addr6);
#endif
      addr6.sin6_port = htons (allPorts[i]);
      
      if(bind(sock, (struct sockaddr*)&addr6, sizeof(addr6)))
      {
	syslog (LOG_ERR, "bind: %m");
	close(sock);
      }
      else if(listen(sock, 128))
      {
	syslog (LOG_ERR, "listen: %m");
	close(sock);
      }
      else
	add_read_fd(sock, connection_accepter, NULL);
    }
  }
  
  return 0;
}

void connection_accepter(int sock, void *user, int urgent)
{
  int csock = accept(sock, NULL, NULL);
  int i;
  
  if(csock >= 0)
  {
    switch(forkConnections)
    {
    case 0:
      break;
    case 1:
      switch(fork())
      {
      case -1:
	syslog (LOG_ERR, "fork: %m");
	return;
      case 0:
	break;
      default:
	close(csock);
	return;
      }
      dup2(csock, 0);
      dup2(csock, 1);
      closelog ();
      csock = 0;
      for(i = 3; i < getdtablesize(); i++)
      {
	remove_read_fd(i);
	remove_write_fd(i);
	close(i);
      }
      openlog (getprogname (), LOG_PID | LOG_NDELAY | (debug? LOG_PERROR : 0),
	    LOG_FTP);
      break;
    case 2:
      /* TODO vfork()/exec() */
      return;
    }
    new_connection (csock, user);
  }
  else
    syslog (LOG_ERR, "accept: %m");
}
