/*****************************************************************************\
* Copyright (c) 2002 Pelle Johansson.                                         *
* All rights reserved.                                                        *
*                                                                             *
* This file is part of the moftpd package. Use and distribution of            *
* this software is governed by the terms in the file LICENCE, which           *
* should have come with this package.                                         *
\*****************************************************************************/

/* $moftpd: user.c,v 1.14.2.2 2003/09/25 23:58:40 morth Exp $ */
/* $Id$ */

#include "user.h"

#include "main.h"
#include "utf8fs/file.h"
#include "utf8fs/memory.h"

extern int forkConnections;

user_t *new_user(const char *name, void *parent)
{
  user_t *res = palloc (sizeof (user_t), parent);
  if (!res)
    return NULL;
  
  if (name)
    res->name = pstring (name, res);
  
  return res;
}

user_t *copy_user(const user_t *u, const char *newname, void *newparent)
{
  user_t *res = palloc (sizeof(user_t), newparent);
  
  if(!res)
    return NULL;
  
  memcpy(res, u, sizeof(user_t));
  if (newname)
    res->name = pstring (newname, res);
  else
    res->name = pattach (u->name, res);
  res->password = pattach (u->password, res);
  res->home = pattach (u->home, res);
  res->chroot = pattach (u->chroot, res);
  res->next = NULL;
  
  return res;
}

user_t *find_passwd_user(const char *name, void *newparent)
{
#ifdef HAVE_GETSPNAM
  struct spwd *spwd;
#endif
  struct passwd *pwd;
  struct group *gr;
  user_t *res;
  char **gusr;
  
#ifdef HAVE_GETSPNAM
  spwd = getspnam (name);
#endif
  pwd = getpwnam (name);
  if (!pwd)
    return NULL;
  
  res = palloc(sizeof(user_t), newparent);
  if(!res)
  {
    syslog (LOG_ERR, "palloc: %m");
    return NULL;
  }
  
  res->name = pstring (name, res);
  if(!res->name)
  {
    syslog (LOG_ERR, "pstring: %m");
    pfree(res, newparent);
    return NULL;
  }
  
#ifdef HAVE_GETSPNAM
  if (spwd)
    res->password = pstring (spwd->sp_pwdp, res);
  else
#endif
    res->password = pstring (pwd->pw_passwd, res);
  res->home = pstring (pwd->pw_dir, res);
  res->uid = pwd->pw_uid;
  res->gid = pwd->pw_gid;
  
  res->ngroups = 0;
  setgrent();
  while((gr = getgrent()) && res->ngroups < NGROUPS_MAX)
  {
    for(gusr = gr->gr_mem; *gusr; gusr++)
      if(!strcmp(*gusr, name))
	res->groups[res->ngroups++] = gr->gr_gid;
  }
  endgrent();
  
  res->allowLogin = 1;
  res->allowSecLogin = 1;
  res->anonymous = 0;
  res->passwordNeeded = strlen(res->password) > 0;
  res->chroot = NULL;
  res->next = NULL;
  
  return res;
}

user_t *find_server_user (const server_t *server, const char *name, void *newparent)
{
  user_t *res;
  
  for (res = server->users; res; res = res->next)
    if (!strcmp (name, res->name))
      return copy_user (res, NULL, newparent);
  return NULL;
}

user_t *find_user(const server_t *server, const char *name, int allowNoLogin, void *newparent)
{
  user_t *res;
  user_t *suser;
  
  name = server_expand_alias(server, name);
  
  res = find_passwd_user(name, newparent);
  suser = find_server_user(server, name, newparent);
  
  if(!res && !suser)
    return NULL;
  
  if(res)
  {
    res->allowLogin = server->allowLogin;
    if(server->chroot)
      res->chroot = pattach (server->chroot, res);
    if(!server->passwordNeeded)
      res->passwordNeeded = 0;
    res->access = server->access;
    res->maxIdle = server->maxIdle;
  }
  
  if(!res)
    res = suser;
  else if(suser)
  {
    if(suser->spec_flags & ufPassword)
    {
      pfree (res->password, res);
      res->password = pattach (suser->password, res);
    }
    if(suser->spec_flags & ufHome)
    {
      pfree (res->home, res);
      res->home = pattach (suser->home, res);
    }
    if(suser->spec_flags & ufUid)
      res->uid = suser->uid;
    if(suser->spec_flags & ufGid)
      res->gid = suser->gid;
    if(suser->spec_flags & (ufUid | ufGid))
      res->ngroups = 0;
    if(suser->spec_flags & ufAnonymous)
      res->anonymous = suser->anonymous;
    if(suser->spec_flags & ufChroot)
    {
      pfree (res->chroot, res);
      res->chroot = pattach (suser->chroot, res);
    }
    res->access = pattach (suser->access, res);
    res->allowLogin = suser->allowLogin;
    res->maxIdle = suser->maxIdle;
    res->passwordNeeded = suser->passwordNeeded;
    pfree (suser, newparent);
  }
  
  if(!allowNoLogin && !res->allowLogin)
  {
    pfree(res, newparent);
    return NULL;
  }
  
  if (res->chroot)
  {
    char *tmp = res->chroot;
    res->chroot = pstring (full_path (tmp, res->home), res);
    pfree (tmp, res);
  }
  
  return res;
}

const char *user_pass(const user_t *user)
{
  if(!user)
    return NULL;
  return user->password;
}

void user_setup_environ(const user_t *user, int authed)
{
  if(user->chroot && authed)
    set_root(user->chroot);
  else
    set_root("/");
  set_gid(user->gid, authed && forkConnections);
  set_groups(user->ngroups, user->groups);
  set_uid(user->uid, authed && forkConnections);
  set_access_chain(user->access);
}
