/*****************************************************************************\
* Copyright (c) 2002 Pelle Johansson.                                         *
* All rights reserved.                                                        *
*                                                                             *
* This file is part of the moftpd package. Use and distribution of            *
* this software is governed by the terms in the file LICENCE, which           *
* should have come with this package.                                         *
\*****************************************************************************/

/* $moftpd: memory.c,v 1.3.2.2 2004/03/16 20:07:31 morth Exp $ */
/* $Id$ */

#include "memory.h"

#include "system.h"

#ifdef DEBUG
extern int debug;
static void DOBUG (const char *fmt, ...)
{
  va_list ap;
  
  va_start (ap, fmt);
  vsyslog (LOG_DEBUG, fmt, ap);
  va_end (ap);
}
#undef DEBUG
#define DEBUG(level, msg) if (debug >= level) DOBUG msg
#define DEXPR(x) x
#else
#define DEBUG(level, msg)
#define DEXPR(x)
#endif

#define stringHashSize 31 // gcc chokes on const int

static memory_t *tMem;
static memdata_t *stringHash[stringHashSize];
static unsigned char randVals[256];
static int sweepMark;
static memdata_t *memChain;

void mem_init (void)
{
  int fd = open ("/dev/urandom", O_RDONLY);
  int i;
  
  if (fd >= 0)
  {
    read (fd, randVals, 256);
    close (fd);
  }
  else
  {
    srand (time (NULL));
    for (i = 0; i < 256; i++)
      randVals[i] = random ();
  }
}

void *talloc(size_t sz)
{
  memory_t *mem = calloc(1, sz + sizeof(memory_t) * pNumTempLinks);
  
  if(!mem)
    return NULL;
  
  mem[pNextTemp].link = tMem;
  tMem = mem;
  
  DEBUG (2, ("talloc: %p", mem + pNumTempLinks));
  
  return mem + pNumTempLinks;
}

void *trealloc(void *ptr, size_t sz)
{
  memory_t *mem = (memory_t*)ptr - pNumTempLinks, *res, *pmem, *nmem;
  
  if (!ptr)
    return talloc (sz);
  
  nmem = mem[pNextTemp].link;
  res = realloc(mem, sz + sizeof(memory_t) * pNumTempLinks);
  
  if (res == mem)
    return res + pNumTempLinks;
  
  DEBUG (2, ("trealloc: %p -> %p", mem + pNumTempLinks, res + pNumTempLinks));
  
  if(mem == tMem)
  {
    if(res)
      tMem = res;
    else
      tMem = nmem;
  }
  else
  {
    for (pmem = tMem; pmem[pNextTemp].link && pmem[pNextTemp].link != mem;
	 pmem = pmem->link)
      ;
    if (pmem[pNextTemp].link)
    {
      if(res)
	pmem[pNextTemp].link = res;
      else
	pmem[pNextTemp].link = nmem;
    }
  }
  
  if(res)
    return res + pNumTempLinks;
  return NULL;
}

char *tstring(const char *str)
{
  char *res = talloc (strlen (str) + 1);
  
  if(res)
    strcpy(res, str);
  
  return res;
}

static void mark_memory (memdata_t *mem)
{
  memory_t *cmem;
  
  if (mem->mark == sweepMark)
    return;
  mem->mark = sweepMark;
  if (mem->isHashed)
    return;
  for (cmem = mem->firstChild.link; cmem; cmem = cmem[pNextChild].link)
    mark_memory (cmem[pChild].data);
}

void tfree_all(void)
{
  memory_t *nmem;
  memdata_t *chain, *pchain;
  
  while(tMem)
  {
    nmem = tMem[pNextTemp].link;
    DEBUG (2, ("tfree: %p", tMem + pNumTempLinks));
    free(tMem);
    tMem = nmem;
  }
  
  sweepMark ^= 1;
  pchain = NULL;
  for (chain = memChain; chain && !chain->firstParent.link; chain = chain->next)
    mark_memory (pchain = chain);
  chain = memChain;
  while (chain)
  {
    if (chain->mark != sweepMark)
    {
      pchain->next = chain->next;
      DEBUG (2, ("collected: %p", chain + 1));
      pfree (chain + 1, (void*)-1);
      chain = pchain->next;
    }
    else
    {
      pchain = chain;
      chain = chain->next;
    }
  }
}

void *proot(void)
{
  return palloc(1, NULL);
}

void *palloc(size_t sz, void *parent)
{
  memdata_t *mem = calloc (1, sz + sizeof (memdata_t));
  
  if(!mem)
    return NULL;
  
  mem->mark = sweepMark;
  mem->next = memChain;
  memChain = mem;
  
  DEBUG (2, ("palloc: %p", mem + 1));
  
  if(!parent)
    return mem + 1;
  
  return pattach(mem + 1, parent);
}

void *prealloc(void *ptr, size_t sz)
{
  memdata_t *mem = (memdata_t*)ptr - 1, *res;
  memory_t *plink = mem->firstParent.link, *clink = mem->firstChild.link,
    *nlink;
  memdata_t *chain;
  
  res = realloc (mem, sz + sizeof (memdata_t));
  
  if (res == mem)
    return res + 1;
  
  DEBUG (2, ("prealloc: %p -> %p", mem + 1, res + 1));
  
  if (mem == memChain)
    memChain = res;
  else
  {
    for (chain = memChain; DEXPR (chain &&) chain->next != mem; chain = chain->next)
      ;
#ifdef DEBUG
    if (!chain)
      DEBUG (1, ("prealloc: %p not found in chain.", mem + 1));
#endif
    chain->next = res;
  }
  while(clink)
  {
    nlink = clink[pNextChild].link;
    if(res)
      clink[pParent].data = res;
    else
      pfree(clink[pChild].data + 1, ptr);
    clink = nlink;
  }
  while(plink)
  {
    nlink = plink[pNextParent].link;
    if(res)
      plink[pChild].data = res;
    else
    {
      if (plink[pParent].data->firstChild.link == plink)
	plink[pParent].data->firstChild.link = plink[pNextChild].link;
      else
      {
	for (clink = plink[pParent].data->firstChild.link;
	     clink[pNextChild].link != plink; clink = clink[pNextChild].link)
	  ;
	clink[pNextChild].link = plink[pNextChild].link;
      }
      DEBUG (2, ("prealloc: unlink %p -> %p", plink[pParent], plink[pChild]));
      free (plink);
    }
    plink = nlink;
  }
  
  if(res)
    return res + 1;
  return NULL;
}

void *pattach(void *child, void *parent)
{
  memdata_t *pmem = (memdata_t*)parent - 1;
  memdata_t *cmem = (memdata_t*)child - 1;
  memory_t *mem;
  memdata_t *chain;
  
  if(!child)
    return NULL;
  if (!parent)
    return child;
#ifdef DEBUG
  if (pmem->isHashed)
  {
    DEBUG (1, ("pattach: parent %p is hashed (child %p)", parent, child));
    return NULL;
  }
  
  for(mem = cmem->firstParent.link; mem; mem = mem[pNextParent].link)
  {
    if(mem[pParent].data == pmem)
      DEBUG (1, ("pattach: %p -> %p already linked.", parent, child));
  }
#endif
  
  mem = malloc(sizeof(memory_t) * pNumLinkLinks);
  if(mem)
  {
    DEBUG (2, ("pattach: linking %p -> %p", parent, child));
    mem[pNextParent].link = cmem->firstParent.link;
    mem[pNextChild].link = pmem->firstChild.link;
    mem[pParent].data = pmem;
    mem[pChild].data = cmem;
    cmem->firstParent.link = mem;
    pmem->firstChild.link = mem;
  }
  else
    DEBUG (1, ("pattach: malloc failed for %p -> %p: %m", parent, child));
  
  // Keep roots at top of chain.
  if (cmem->next && !cmem->next->firstParent.link)
  {
    DEBUG (2, ("pattach: Unrooting %p", child));
    if (cmem == memChain)
      chain = NULL;
    else
    {
      for (chain = memChain; DEXPR (chain &&) chain->next != cmem; chain = chain->next)
	;
#ifdef DEBUG
      if (!chain)
	DEBUG (1, ("pattach: %p not found in chain.", child));
#endif
    }
    while (cmem->next && !cmem->next->firstParent.link)
    {
      if (chain)
	chain = chain->next = cmem->next;
      else
	chain = memChain = cmem->next;
      cmem->next = chain->next;
      chain->next = cmem;
    }
  }
  
  return child;
}

void *padopt(void *child, void *oldparent, void *newparent)
{
  memdata_t *opmem = (memdata_t*)oldparent - 1;
  memdata_t *npmem = (memdata_t*)newparent - 1;
  memdata_t *cmem = (memdata_t*)child - 1;
  memory_t *mem, *ccmem = NULL, *pcmem, *ppmem;
  int i;
  
  if (!child)
    return NULL;
  if (!newparent || npmem->isHashed)
    return child;
  
  pcmem = NULL;
  i = 0;
  for(mem = cmem->firstParent.link; mem && i < 3; mem = mem[pNextParent].link)
  {
    if(!(i & 1))
      ccmem = mem;
    if(mem[pParent].data == npmem)
      i |= 2;
    else if(mem[pParent].data == opmem)
      i |= 1;
    if(!(i & 1))
      pcmem = mem;
  }
  
  switch(i)
  {
  case 0:
    DEBUG (1, ("padopt: No old link %p -> %p.", oldparent, child));
    return pattach (child, newparent);
  case 1:
    /* This is the one we _should_ get. */
    DEBUG (2, ("padopt: %p adopting %p from %p.", newparent, child, oldparent));
    if(ccmem == opmem->firstChild.link)
      opmem->firstChild.link = ccmem[pNextChild].link;
    else
    {
      for(ppmem = opmem->firstChild.link; DEXPR (ppmem[pNextChild].link &&)
		ppmem[pNextChild].link != ccmem; ppmem = ppmem[pNextChild].link)
	;
#ifdef DEBUG
      if(!ppmem[pNextChild].link)
      {
	DEBUG (1, ("padopt: Couldn't find link to child in old parent."));
	return child;
      }
#endif
      ppmem[pNextChild].link = ccmem[pNextChild].link;
    }
    ccmem[pNextChild].link = npmem->firstChild.link;
    ccmem[pParent].data = npmem;
    npmem->firstChild.link = ccmem;
    return child;
  case 3:
    DEBUG (1, ("padopt: Both %p and %p already linked to %p.", oldparent, newparent, child));
    if(ccmem == opmem->firstChild.link)
      opmem->firstChild.link = ccmem[pNextChild].link;
    else
    {
      for(ppmem = opmem->firstChild.link; DEXPR (ppmem[pNextChild].link &&)
		ppmem[pNextChild].link != ccmem; ppmem = ppmem[pNextChild].link)
	;
#ifdef DEBUG
      if(!ppmem[pNextChild].link)
      {
	DEBUG (1, ("padopt: Couldn't find old link."));
	return child;
      }
#endif
      ppmem[pNextChild].link = ccmem[pNextChild].link;
    }
    
    if(pcmem)
      pcmem[pNextParent].link = ccmem[pNextParent].link;
    else
      cmem->firstParent.link = ccmem[pNextParent].link;
    
    DEBUG (2, ("padopt: Unlinking %p -> %p.", oldparent, child));
    free(ccmem);
    return child;
  default:
    /*
     * Everything ok already. Except that the call was made
     * erronously that is.
     */
    DEBUG (1, ("padopt: %p is already parent of %p and %p isn't.", newparent, child, oldparent));
    return child;
  }
}

void pfree(void *child, void *parent)
{
  memdata_t *cmem = (memdata_t*)child - 1;
  memdata_t *pmem = (memdata_t*)parent - 1;
  memory_t *mem, *pcmem, *ppmem;
  memhash_t *hash, *phash;
  memdata_t *chain;
  
  if(!child)
    return;
  
  pcmem = NULL;
  if (parent && parent != (void*)-1)
  {
#ifdef DEBUG
    if (pmem->isHashed)
    {
      DEBUG (1, ("pfree: Parent %p is hashed (child %p).", parent, child));
      return;
    }
#endif
    
    for (mem = cmem->firstParent.link; mem; mem = mem[pNextParent].link)
    {
      if(mem[pParent].data == pmem)
	break;
      pcmem = mem;
    }
    if (!mem)
    {
#ifdef DEBUG
      DEBUG (1, ("pfree: Parent %p not found in child %p.", parent, child));
      ppmem = NULL;
      for (mem = pmem->firstChild.link; mem; mem = mem[pNextChild].link)
      {
	if (mem[pChild].data == cmem)
	  break;
	ppmem = mem;
      }
      if (mem)
      {
	DEBUG (1, ("pfree: but child found in parent."));
	ppmem[pNextChild].link = mem[pNextChild].link;
	free (mem);
      }
#endif
      return;
    }
    
    if (mem == pmem->firstChild.link)
      pmem->firstChild.link = mem[pNextChild].link;
    else
    {
      for (ppmem = pmem->firstChild.link; DEXPR (ppmem[pNextChild].link &&)
	     ppmem[pNextChild].link != mem; ppmem = ppmem[pNextChild].link)
	;
#ifdef DEBUG
      if (!ppmem[pNextChild].link)
	DEBUG (1, ("pfree: Child %p not found in parent %p.", child, parent));
#endif
      ppmem[pNextChild].link = mem[pNextChild].link;
    }
    if (pcmem)
      pcmem[pNextParent].link = mem[pNextParent].link;
    else
      cmem->firstParent.link = mem[pNextParent].link;
    
    DEBUG (2, ("pfree: Unlinking %p -> %p.", parent, child));
    free(mem);
  }
  else
  {
    DEBUG (2, ("pfree: Really freeing %p.", child));
    if (cmem->isHashed)
    {
      hash = cmem->firstChild.hash;
      if (*hash->bucket && (*hash->bucket)->firstChild.hash == hash)
	*hash->bucket = hash->next;
      else
      {
	for (phash = (*hash->bucket)->firstChild.hash; phash->next &&
	       phash->next->firstChild.hash != hash;
	     phash = phash->next->firstChild.hash)
	  ;
	if (phash)
	  phash->next = hash->next;
	free (cmem->firstChild.hash);
      }
    }
    else
    {
      while(cmem->firstChild.link)
	pfree (cmem->firstChild.link[pChild].data + 1, child);
    }
    while (cmem->firstParent.link)
      pfree (child, cmem->firstParent.link[pParent].data + 1);
    if (parent != (void*)-1)
    {
      if (cmem == memChain)
	memChain = cmem->next;
      else
      {
	for (chain = memChain; DEXPR (chain &&) chain->next != cmem; chain = chain->next)
	  ;
#ifdef DEBUG
	if (!chain)
	  DEBUG (1, ("%p not found in chain.", child));
#endif
	chain->next = cmem->next;
      }
    }
    free(cmem);
  }
}

void *pchild(const void *parent, const void *prevChild)
{
  const memdata_t *const pmem = (memdata_t*)parent - 1;
  
  if (!parent)
    return NULL;
  
  if(prevChild)
  {
    const memdata_t *const cmem = (memdata_t*)prevChild - 1;
    const memory_t *mem;
    
    /*
     * Most nodes tend to have more children
     * than parents, so usually it should be faster to search for the
     * parent in the child's list.
     */
    for(mem = cmem->firstParent.link; mem; mem = mem[pNextParent].link)
      if(mem[pParent].data == pmem)
      {
	if(mem[pNextChild].link)
	  return mem[pNextChild].link[pChild].data + 1;
	return NULL;
      }
  }
  else if(pmem->firstChild.link)
    return pmem->firstChild.link[pChild].data + 1;
  return NULL;
}

char *pstring(const char *str, void *parent)
{
  int hval = 0;
  const unsigned char *sp = (unsigned char*)str;
  memdata_t *mem;
  char *res;
  
  if (!str)
    return NULL;
  
  while (*sp)
  {
    hval += randVals[(*sp + ((char*)sp - str)) & 0xFF];
    sp++;
  }
  hval %= stringHashSize;
  
  mem = stringHash[hval];
  while (mem)
  {
    if (!strcmp ((char*)(mem + 1), str))
      return (char*)pattach (mem + 1, parent);
    mem = mem->firstChild.hash->next;
  }
  
  res = palloc (strlen (str) + 1, parent);
  
  if (res)
  {
    strcpy (res, str);
    mem = (memdata_t*)res - 1;
    mem->firstChild.hash = malloc (sizeof (memhash_t));
    if (mem->firstChild.hash)
    {
      mem->isHashed = 1;
      mem->firstChild.hash->bucket = &stringHash[hval];
      mem->firstChild.hash->next = stringHash[hval];
      stringHash[hval] = mem;
    }
  }
  
  return res;
}
